### configuration file for odf_process_all.py
#
# TODO: organize these by editable/fixed variables
from importlib import resources
import yaml

with resources.open_text("ctdcal", "user_settings.yaml") as f:
    settings = yaml.safe_load(f)

# Unpack user settings (any sanitizing/checks needed? probably)
expocode = settings["expocode"]
section_id = settings["section_id"]
ctd_serial = settings["ctd_serial"]
ctd_outputs = settings["ctd_outputs"]

# CTD file (.ct1) variable outputs
# move elsewhere when xarray is implemented
# TODO: check against cchdo.params?
ctd_col_names, ctd_col_units = [], []
for (param, attrs) in ctd_outputs.items():
    if param == "CTDPRS":
        ctd_col_names += [param]
        ctd_col_units += [attrs["units"]]
    else:
        ctd_col_names += [param, f"{param}_FLAG_W"]
        ctd_col_units += [attrs["units"], ""]

# List of directories for I/O purposes
dirs = {
    "ssscc": "data/ssscc/",
    "raw": "data/raw/",
    "converted": "data/converted/",
    "time": "data/time/",
    "pressure": "data/pressure/",
    "bottle": "data/bottle/",
    "salt": "data/salt/",
    "reft": "data/reft/",
    "oxygen": "data/oxygen/",
    "logs": "data/logs/",
}
fig_dirs = {
    "t1": "data/logs/fitting_figs/temp_primary/",
    "t2": "data/logs/fitting_figs/temp_secondary/",
    "c1": "data/logs/fitting_figs/cond_primary/",
    "c2": "data/logs/fitting_figs/cond_secondary/",
    "ox": "data/logs/fitting_figs/oxy_primary/",
    "rinko": "data/logs/fitting_figs/oxy_rinko/",
}

# remnant of old system, will be pushed into xarray metadata/attrs
# Labels for CTD columns
column = {
    "p": "CTDPRS",
    "t1": "CTDTMP1",
    "t2": "CTDTMP2",
    "c1": "CTDCOND1",
    "c2": "CTDCOND2",
    "sal": "CTDSAL",
    # "s1": "CTDSAL1",  # TODO: calc salinity from primary and secondary sensors
    # "s2": "CTDSAL2",
    "rinko_oxy": "FREE1",  # CHECK THIS!
    "oxyvolts": "CTDOXYVOLTS",
    "refT": "T90",
    "refC": "BTLCOND",
    "refS": "SALNTY",
    "refO": "OXYGEN",
    "lat": "GPSLAT",
    "lon": "GPSLON",
}

# List of columns to filter
filter_cols = []
for x in ["p", "t1", "t2", "c1", "c2", "sal", "rinko_oxy", "oxyvolts", "lat", "lon"]:
    filter_cols.append(column[x])



"""Needs truncation implemented."""

"""Following equations taken from SBE3+ calib sheet"""

def temp_its90(g, h, i, j, f0, f):
    """SBE equation for converting engineering units to Celcius according to ITS-90.

    g, h, i, j, f0: coefficients given in calibration.
    f: frequency sampled by sensor.

    Original form from calib sheet dated 2012:
    Temperature ITS-90 = 1/{g+h[ln(f0/f )]+i[ln2(f0/f)]+j[ln3(f0/f)]} - 273.15 (°C)

    """
    ITS90 = 1/(g +
        h * (math.log(f0/f)) +
        i * math.pow((math.log(f0/f)),2) +
        j * math.pow((math.log(f0/f)),3)
        ) - 273.15
    return ITS90

def temp_its68(a, b, c, d, f0, f):
    """SBE equation for converting engineering units to Celcius according to IPTS-68.

    a, b, c, d, f0: coefficients given in calibration.
    f: frequency sampled by sensor.

    Original form from calib sheet dated 2012:
    Temperature IPTS-68 = 1/{a+b[ln(f0/f )]+c[ln2(f0/f)]+d[ln3(f0/f)]} - 273.15 (°C)

    """
    ITS68 = 1/(a +
        b * (math.log(f0/f)) +
        c * math.pow((math.log(f0/f)),2) +
        d * math.pow((math.log(f0/f)),3)
        ) - 273.15
    return ITS68

"""Following equations taken from SBE 43 calib sheet."""

def oxy(Soc, Voffset, Tau20, A, B, C, E, P, K, T, S, V):
    """SBE equation for converting engineering units to oxygen (ml/l).

    asdasdas

    Original equation from calib sheet dated 2014:
    Oxygen (ml/l) = Soc * (V + Voffset) * (1.0 + A * T + B * T + C * T ) * OxSol(T,S) * exp(E * P / K)

    """

    oxygen = Soc * (V + Voffset) * (1.0 + A * T + B * T + C * T ) * OxSol(T,S) * math.exp(E * P / K)
    return oxygen

def OxSol(T,S):
    """Eq. 8 from Garcia and Gordon, 1992.

    Inputs:
    T = ITS-90 Temperature
    S = Practical Salinity
    """

    x = S
    y = math.log((298.15 - T)/(273.15 + T))

    """umol/kg coefficients
    a0 =  5.80871
    a1 =  3.20291
    a2 =  4.17887
    a3 =  5.10006
    a4 = -9.86643e-2
    a5 =  3.80369
    b0 = -7.01577e-3
    b1 = -7.70028e-3
    b2 = -1.13864e-2
    b3 = -9.51519e-3
    c0 = -2.75915e-7

    """

    """ml/l coefficients"""
    a0 = 2.00907
    a1 = 3.22014
    a2 = 4.05010
    a3 = 4.94457
    a4 = -2.56847e-1
    a5 = 3.88767
    b0 = -6.24523e-3
    b1 = -7.37614e-3
    b2 = -1.03410e-2
    b3 = -8.17083e-3
    c0 = -4.88682e-7

    O2sol = math.exp(a0 + y*(a1 + y*(a2 + y*(a3 + y*(a4 + a5*y)))) + x*(b0 + y*(b1 + y*(b2 + b3*y)) + c0*x))
    return O2sol

def Cond(g, h, i, j, CPcor, CTcor, F, t, p):
    """SBE equation for converting frequency to conductivity.

    Inputs:
    g: coefficient
    h: coefficient
    i: coefficient
    j: coefficient
    CPcor: coefficient (nominal)
    CTcor: coefficient (nominal)

    F: instrument frequency
    t: temperature (ITS-90 degrees C)
    p: pressure (decibars)
    """
    f = F/1000
    Conductivity = (g + h * round(math.pow(f,2),3) + i * round(math.pow(f,3),3) + j * round(math.pow(f,4),3) ) / (1 + CTcor * t + CPcor * p)
    return round(Conductivity,5)



"""
Library and scripts for processing CTD profiles with bottle
data, and similar data types such as ARGO and moored CTD
sensors.
"""

import logging
from importlib import resources
from importlib.metadata import version, PackageNotFoundError

log = logging.getLogger(__name__)

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass


def get_ctdcal_config():
    """
    Find and load config file from ctdcal module folder (ctdcal/ctdcal/).

    File read mimicked from pallets/flask
    """
    # compile config.py and save variables to dict
    config = {}
    with resources.path("ctdcal", "config.py") as filepath:
        try:
            with open(filepath, mode="rb") as f:
                exec(compile(f.read(), filepath, "exec"), config)
        except OSError:
            log.error(f"Failed to load config file {filepath}")

    for k in list(config.keys()):
        if k.startswith("__"):
            del config[k]

    return type("config", (object,), config)  # make into class for easier calls



#!/usr/bin/env python
import datetime
from pathlib import Path

import numpy as np
import pandas as pd


def report_pressure_details(stacast, log_file, start, end):
    """
    Write start/end deck pressure to ondeck_pressure.csv log file.

    Parameters
    ----------
    stacast : str
        station cast data for file
    c_file : str
        file name location for cast details
    start : str
        cast start time from top of cast after 10m soak
    end : str
        cast end time when instrument leaves water

    Returns
    -------
    None
    """
    df = pd.DataFrame(
        {"SSSCC": stacast, "ondeck_start_p": start, "ondeck_end_p": end}, index=[0]
    )
    add_header = not Path(log_file).exists()  # add header iff file doesn't exist
    with open(log_file, "a") as f:
        df.to_csv(f, mode="a", header=add_header, index=False)

    return True


def report_cast_details(
    stacast, c_file, start, end, bottom, start_p, max_p, b_alt, b_lat, b_lon
):
    """
    Write cast details to cast_details.csv log file.

    Parameters
    ----------
    stacast : str
        station cast data for file
    c_file : str
        file name location for cast details
    start : str
        cast start time from top of cast after 10m soak
    end : str
        cast end time when instrument leaves water
    bottom : str
        bottom of cast time when instrument reaches max depth
    start_p : str
        starting pressure at the time the cast begins
    max_p : str
        maximum pressure for entire cast
    b_alt : str
        altimeter value at bottom of cast
    b_lat : str
        latitude at bottom of cast
    b_lon : str
        longitude at bottom of cast

    Returns
    -------
    None
    """
    df = pd.DataFrame(
        {
            "SSSCC": stacast,
            "start_time": start,
            "bottom_time": bottom,
            "end_time": end,
            "start_pressure": start_p,
            "max_pressure": max_p,
            "altimeter_bottom": b_alt,
            "latitude": b_lat,
            "longitude": b_lon,
        },
        index=[0],
    )
    add_header = not Path(c_file).exists()  # add header iff file doesn't exist
    with open(c_file, "a") as f:
        df.to_csv(f, mode="a", header=add_header, index=False)

    return



"""Library to create SBE .btl equivalent files.
TODO: allow for variable bottle fire scans instead of SBE standard 36
    ex: user doesn't know how to change the config for the cast to add more scans,
    instead does it post-cast?

Joseph Gum SIO/ODF
Nov 7, 2016
"""

import csv
import logging
import statistics
import sys
from collections import OrderedDict
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd

from . import flagging as flagging
from . import get_ctdcal_config
from . import oxy_fitting as oxy_fitting

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)

BOTTLE_FIRE_COL = "btl_fire"
BOTTLE_FIRE_NUM_COL = "btl_fire_num"


# Retrieve the bottle data from a converted file.
def retrieveBottleDataFromFile(converted_file):

    converted_df = pd.read_pickle(converted_file)

    return retrieveBottleData(converted_df)


# Retrieve the bottle data from a dataframe created from a converted file.
def retrieveBottleData(converted_df):
    if BOTTLE_FIRE_COL in converted_df.columns:
        converted_df[BOTTLE_FIRE_NUM_COL] = (
            (
                (converted_df[BOTTLE_FIRE_COL])
                & (
                    converted_df[BOTTLE_FIRE_COL]
                    != converted_df[BOTTLE_FIRE_COL].shift(1)
                )
            )
            .astype(int)
            .cumsum()
        )
        # converted_df['bottle_fire_num'] = ((converted_df[BOTTLE_FIRE_COL] == False)).astype(int).cumsum()
        return converted_df.loc[converted_df[BOTTLE_FIRE_COL]]
        # return converted_df
    else:
        log.error("Bottle fire column:", BOTTLE_FIRE_COL, "not found")

    return pd.DataFrame()  # empty dataframe


def bottle_mean(btl_df):
    """Compute the mean for each bottle from a dataframe."""
    btl_max = int(btl_df[BOTTLE_FIRE_NUM_COL].tail(n=1))
    i = 1
    output = pd.DataFrame()
    while i <= btl_max:
        output = pd.concat(
            (
                output,
                btl_df[btl_df[BOTTLE_FIRE_NUM_COL] == i]
                .mean()
                .to_frame(name=i)
                .transpose(),
            )
        )
        i += 1
    return output


def bottle_median(btl_df):
    """Compute the median for each bottle from a dataframe."""
    btl_max = int(btl_df[BOTTLE_FIRE_NUM_COL].tail(n=1))
    i = 1
    output = pd.DataFrame()
    while i <= btl_max:
        output = pd.concat(
            (
                output,
                btl_df[btl_df[BOTTLE_FIRE_NUM_COL] == i]
                .median()
                .to_frame(name=i)
                .transpose(),
            )
        )
        i += 1
    return output


def _load_btl_data(btl_file, cols=None):
    """
    Loads "bottle mean" CTD data from .pkl file. Function will return all data unless
    cols is specified (as a list of column names)
    """

    btl_data = pd.read_pickle(btl_file)
    if cols is not None:
        btl_data = btl_data[cols]
    btl_data["SSSCC"] = Path(btl_file).stem.split("_")[0]

    return btl_data


def _load_reft_data(reft_file, index_name="btl_fire_num"):
    """
    Loads reft_file to dataframe and reindexes to match bottle data dataframe
    """
    reft_data = pd.read_csv(reft_file, usecols=["btl_fire_num", "T90", "REFTMP_FLAG_W"])
    reft_data.set_index(index_name)
    reft_data["SSSCC_TEMP"] = Path(reft_file).stem.split("_")[0]
    reft_data["REFTMP"] = reft_data["T90"]

    return reft_data


def _load_salt_data(salt_file, index_name="SAMPNO"):
    """
    Loads salt_file to dataframe and reindexes to match bottle data dataframe
    """
    salt_data = pd.read_csv(
        salt_file, usecols=["SAMPNO", "SALNTY", "BathTEMP", "CRavg"]
    )
    salt_data.set_index(index_name)
    salt_data["SSSCC_SALT"] = Path(salt_file).stem.split("_")[0]
    salt_data.rename(columns={"SAMPNO": "SAMPNO_SALT"}, inplace=True)

    return salt_data


def _add_btl_bottom_data(df, cast, lat_col="LATITUDE", lon_col="LONGITUDE", decimals=4):
    cast_details = pd.read_csv(
        # cfg.dirs["logs"] + "cast_details.csv", dtype={"SSSCC": str}
        cfg.dirs["logs"] + "bottom_bottle_details.csv",
        dtype={"SSSCC": str},
    )
    cast_details = cast_details[cast_details["SSSCC"] == cast]
    # df[lat_col] = np.round(cast_details["latitude"].iat[0], decimals)
    # df[lon_col] = np.round(cast_details["longitude"].iat[0], decimals)
    df[lat_col] = cast_details["latitude"].iat[0]
    df[lon_col] = cast_details["longitude"].iat[0]

    ts = pd.to_datetime(cast_details["bottom_time"].iat[0], unit="s")
    date = ts.strftime("%Y%m%d")
    hour = ts.strftime("%H%M")
    df["DATE"] = date
    df["TIME"] = hour
    return df


def load_all_btl_files(ssscc_list, cols=None):
    """
    Load bottle and secondary (e.g. reference temperature, bottle salts, bottle oxygen)
    files for station/cast list and merge into a dataframe.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to load
    cols : list of str, optional
        Subset of columns to load, defaults to loading all

    Returns
    -------
    df_data_all : DataFrame
        Merged dataframe containing all loaded data

    """
    df_data_all = pd.DataFrame()

    for ssscc in ssscc_list:
        log.info("Loading BTL data for station: " + ssscc + "...")
        btl_file = cfg.dirs["bottle"] + ssscc + "_btl_mean.pkl"
        btl_data = _load_btl_data(btl_file, cols)

        ### load REFT data
        reft_file = cfg.dirs["reft"] + ssscc + "_reft.csv"
        try:
            reft_data = _load_reft_data(reft_file)
        except FileNotFoundError:
            log.warning(
                "Missing (or misnamed) REFT Data Station: "
                + ssscc
                + "...filling with NaNs"
            )
            reft_data = pd.DataFrame(index=btl_data.index, columns=["T90"], dtype=float)
            reft_data["btl_fire_num"] = btl_data["btl_fire_num"].astype(int)
            reft_data["SSSCC_TEMP"] = ssscc  # TODO: is this ever used?

        ### load REFC data
        refc_file = cfg.dirs["salt"] + ssscc + "_salts.csv"
        try:
            refc_data = _load_salt_data(refc_file, index_name="SAMPNO")
        except FileNotFoundError:
            log.warning(
                "Missing (or misnamed) REFC Data Station: "
                + ssscc
                + "...filling with NaNs"
            )
            refc_data = pd.DataFrame(
                index=btl_data.index,
                columns=["CRavg", "BathTEMP", "BTLCOND"],
                dtype=float,
            )
            refc_data["SAMPNO_SALT"] = btl_data["btl_fire_num"].astype(int)

        ### load OXY data
        oxy_file = cfg.dirs["oxygen"] + ssscc
        try:
            oxy_data, params = oxy_fitting.load_winkler_oxy(oxy_file)
        except FileNotFoundError:
            log.warning(
                "Missing (or misnamed) REFO Data Station: "
                + ssscc
                + "...filling with NaNs"
            )
            oxy_data = pd.DataFrame(
                index=btl_data.index,
                columns=[
                    "FLASKNO",
                    "TITR_VOL",
                    "TITR_TEMP",
                    "DRAW_TEMP",
                    "TITR_TIME",
                    "END_VOLTS",
                ],
                dtype=float,
            )
            oxy_data["STNNO_OXY"] = ssscc[:3]  # TODO: are these values
            oxy_data["CASTNO_OXY"] = ssscc[3:]  # ever used?
            oxy_data["BOTTLENO_OXY"] = btl_data["btl_fire_num"].astype(int)

        ### clean up dataframe
        # Horizontally concat DFs to have all data in one DF
        btl_data = pd.merge(btl_data, reft_data, on="btl_fire_num", how="outer")
        btl_data = pd.merge(
            btl_data,
            refc_data,
            left_on="btl_fire_num",
            right_on="SAMPNO_SALT",
            how="outer",
        )
        btl_data = pd.merge(
            btl_data,
            oxy_data,
            left_on="btl_fire_num",
            right_on="BOTTLENO_OXY",
            how="outer",
        )

        if len(btl_data) > 36:
            log.error(
                f"""Length of bottle data for {ssscc} is > 36, check for errors in reference parameter files"""
            )

        # Add bottom of cast information (date,time,lat,lon,etc.)
        btl_data = _add_btl_bottom_data(btl_data, ssscc)

        # Merge cast into df_data_all
        try:
            df_data_all = pd.concat([df_data_all, btl_data], sort=False)
        except AssertionError:
            raise AssertionError(
                "Columns of " + ssscc + " do not match those of previous columns"
            )
        # print("* Finished BTL data station: " + ssscc + " *")

    # Drop duplicated columns generated by concatenation
    df_data_all = df_data_all.loc[:, ~df_data_all.columns.duplicated()]

    df_data_all["master_index"] = range(len(df_data_all))

    return df_data_all


def _reft_loader(ssscc, reft_dir):
    # semi-flexible search for reft file (in the form of *ssscc.cap)
    try:
        reft_path = sorted(Path(reft_dir).glob(f"*{ssscc}.cap"))[0]
    except IndexError:
        raise FileNotFoundError

    # this works better than pd.read_csv as format is semi-inconsistent (cf .cap files)
    with open(reft_path, "r", newline="") as f:
        reftF = csv.reader(
            f, delimiter=" ", quoting=csv.QUOTE_NONE, skipinitialspace="True"
        )
        reftArray = []
        for row in reftF:
            if len(row) != 17:  # skip over 'bad' rows (empty lines, comments, etc.)
                continue
            reftArray.append(row)

    reftDF = pd.DataFrame.from_records(reftArray)
    reftDF = reftDF.replace(  # remove text columns, only need numbers and dates
        to_replace=["bn", "diff", "val", "t90", "="], value=np.nan
    )
    reftDF = reftDF.dropna(axis=1)
    reftDF[1] = reftDF[[1, 2, 3, 4]].agg(" ".join, axis=1)  # dd/mm/yy/time cols are
    reftDF.drop(columns=[2, 3, 4], inplace=True)  # read separately; combine into one

    columns = OrderedDict(  # having this as a dict streamlines next steps
        [
            ("index_memory", int),
            ("datetime", object),
            ("btl_fire_num", int),
            ("diff", int),
            ("raw_value", float),
            ("T90", float),
        ]
    )
    reftDF.columns = list(columns.keys())  # name columns
    reftDF = reftDF.astype(columns)  # force dtypes

    # assign initial flags (large "diff" = unstable reading, flag questionable)
    reftDF["REFTMP_FLAG_W"] = 2
    reftDF.loc[reftDF["diff"].abs() >= 3000, "REFTMP_FLAG_W"] = 3

    # add in STNNBR, CASTNO columns
    # TODO: should these be objects or floats? be consistent!
    # string prob better for other sta/cast formats (names, letters, etc.)
    reftDF["STNNBR"] = ssscc[0:3]
    reftDF["CASTNO"] = ssscc[3:5]
    return reftDF


def process_reft(ssscc_list, reft_dir=cfg.dirs["reft"]):
    # TODO: import reft_dir from a config file
    """
    SBE35 reference thermometer processing function. Load in .cap files for given
    station/cast list, perform basic flagging, and export to .csv files.

    Inputs
    ------
    ssscc_list : list of str
        List of stations to process
    reft_dir : str, optional
        Path to folder containing raw salt files (defaults to data/reft/)

    """
    for ssscc in ssscc_list:
        if not Path(reft_dir + ssscc + "_reft.csv").exists():
            try:
                reftDF = _reft_loader(ssscc, reft_dir)
                reftDF.to_csv(reft_dir + ssscc + "_reft.csv", index=False)
            except FileNotFoundError:
                log.warning(
                    "refT file for cast " + ssscc + " does not exist... skipping"
                )
                return


def add_btlnbr_cols(df, btl_num_col):
    df["BTLNBR"] = df[btl_num_col].astype(int)
    # default to everything being good
    df["BTLNBR_FLAG_W"] = 2
    return df


def load_hy_file(path_to_hyfile):
    df = pd.read_csv(path_to_hyfile, comment="#", skiprows=[0])
    df = df[df["EXPOCODE"] != "END_DATA"]
    return df


def export_report_data(df):

    df["STNNBR"] = [int(x[0:3]) for x in df["SSSCC"]]
    df["CTDPRS"] = df["CTDPRS"].round(1)
    cruise_report_cols = [
        "STNNBR",
        "CTDPRS",
        "CTDTMP1",
        "CTDTMP1_FLAG_W",
        "CTDTMP2",
        "CTDTMP2_FLAG_W",
        "REFTMP",
        "CTDCOND1",
        "CTDCOND1_FLAG_W",
        "CTDCOND2",
        "CTDCOND2_FLAG_W",
        "BTLCOND",
        "CTDSAL",
        "CTDSAL_FLAG_W",
        "SALNTY",
        "CTDOXY",
        "CTDOXY_FLAG_W",
        "CTDRINKO",
        "CTDRINKO_FLAG_W",
        "OXYGEN",
    ]

    # add in missing flags
    df["CTDTMP1_FLAG_W"] = flagging.by_residual(
        df["CTDTMP1"], df["REFTMP"], df["CTDPRS"]
    )
    df["CTDTMP2_FLAG_W"] = flagging.by_residual(
        df["CTDTMP1"], df["REFTMP"], df["CTDPRS"]
    )
    df["CTDCOND1_FLAG_W"] = flagging.by_residual(
        df["CTDCOND1"], df["BTLCOND"], df["CTDPRS"]
    )
    df["CTDCOND2_FLAG_W"] = flagging.by_residual(
        df["CTDCOND2"], df["BTLCOND"], df["CTDPRS"]
    )
    df["CTDOXY_FLAG_W"] = flagging.by_percent_diff(df["CTDOXY"], df["OXYGEN"])
    df["CTDRINKO_FLAG_W"] = flagging.by_percent_diff(df["CTDRINKO"], df["OXYGEN"])

    df[cruise_report_cols].to_csv("data/scratch_folder/report_data.csv", index=False)

    return


def export_hy1(df, out_dir=cfg.dirs["pressure"], org="ODF"):
    log.info("Exporting bottle file")
    btl_data = df.copy()
    now = datetime.now()
    file_datetime = now.strftime("%Y%m%d")

    # TODO: move to config; integrate Barna's "params" package instead?
    btl_columns = {
        "EXPOCODE": "",
        "SECT_ID": "",
        "STNNBR": "",
        "CASTNO": "",
        "SAMPNO": "",
        "BTLNBR": "",
        "BTLNBR_FLAG_W": "",
        "DATE": "",
        "TIME": "",
        "LATITUDE": "",
        "LONGITUDE": "",
        "DEPTH": "METERS",
        "CTDPRS": "DBAR",
        "CTDTMP": "ITS-90",
        "CTDSAL": "PSS-78",
        "CTDSAL_FLAG_W": "",
        "SALNTY": "PSS-78",
        "SALNTY_FLAG_W": "",
        # "CTDOXY": "UMOL/KG",
        # "CTDOXY_FLAG_W": "",
        # "CTDRINKO": "UMOL/KG",
        # "CTDRINKO_FLAG_W": "",
        "CTDOXY": "UMOL/KG",
        "CTDOXY_FLAG_W": "",
        "OXYGEN": "UMOL/KG",
        "OXYGEN_FLAG_W": "",
        "REFTMP": "ITS-90",
        "REFTMP_FLAG_W": "",
    }

    # rename outputs as defined in user_settings.yaml
    for param, attrs in cfg.ctd_outputs.items():
        if param not in btl_data.columns:
            btl_data.rename(columns={attrs["sensor"]: param}, inplace=True)

    btl_data["EXPOCODE"] = cfg.expocode
    btl_data["SECT_ID"] = cfg.section_id
    btl_data["STNNBR"] = [int(x[0:3]) for x in btl_data["SSSCC"]]
    btl_data["CASTNO"] = [int(x[3:]) for x in btl_data["SSSCC"]]
    btl_data["SAMPNO"] = btl_data["btl_fire_num"].astype(int)
    btl_data = add_btlnbr_cols(btl_data, btl_num_col="btl_fire_num")

    # sort by decreasing sample number (increasing pressure) and reindex
    btl_data = btl_data.sort_values(
        by=["STNNBR", "SAMPNO"], ascending=[True, False], ignore_index=True
    )

    # switch oxygen primary sensor to rinko
    btl_data["CTDOXY"] = btl_data.loc[:, "CTDRINKO"]
    btl_data["CTDOXY_FLAG_W"] = btl_data.loc[:, "CTDRINKO_FLAG_W"]

    # round data
    # for col in ["CTDTMP", "CTDSAL", "SALNTY", "REFTMP"]:
    #     btl_data[col] = btl_data[col].round(4)
    # for col in ["CTDPRS", "CTDOXY", "OXYGEN"]:
    #     btl_data[col] = btl_data[col].round(1)

    # add depth
    depth_df = pd.read_csv(
        cfg.dirs["logs"] + "depth_log.csv", dtype={"SSSCC": str}, na_values=-999
    ).dropna()
    manual_depth_df = pd.read_csv(
        cfg.dirs["logs"] + "manual_depth_log.csv", dtype={"SSSCC": str}
    )
    full_depth_df = pd.concat([depth_df, manual_depth_df])
    full_depth_df.drop_duplicates(subset="SSSCC", keep="first", inplace=True)
    btl_data["DEPTH"] = -999
    for index, row in full_depth_df.iterrows():
        btl_data.loc[btl_data["SSSCC"] == row["SSSCC"], "DEPTH"] = int(row["DEPTH"])

    # deal with nans
    # TODO: missing REFTMP not obvious til loading data - where to put this?
    # _reft_loader() is not the right place
    # maybe during loading step flag missing OXYGEN, REFTMP, BTLCOND?
    btl_data["REFTMP_FLAG_W"] = flagging.nan_values(
        btl_data["REFTMP_FLAG_W"], old_flags=btl_data["REFTMP_FLAG_W"]
    )
    btl_data = btl_data.where(~btl_data.isnull(), -999)

    # check columns
    try:
        btl_data[btl_columns.keys()]
        # this is lazy, do better
    except KeyError as err:
        log.info("Column names not configured properly... attempting to correct")
        bad_cols = err.args[0].split("'")[1::2]  # every other str is a column name
        for col in bad_cols:
            if col.endswith("FLAG_W"):
                log.warning(col + " missing, flagging with 9s")
                btl_data[col] = 9
            else:
                log.warning(col + " missing, filling with -999s")
                btl_data[col] = -999

    btl_data = btl_data[btl_columns.keys()]
    time_stamp = file_datetime + org
    with open(out_dir + cfg.expocode + "_hy1.csv", mode="w+") as f:
        f.write("BOTTLE, %s\n" % (time_stamp))
        f.write(",".join(btl_columns.keys()) + "\n")
        f.write(",".join(btl_columns.values()) + "\n")
        btl_data.to_csv(f, header=False, index=False)
        f.write("\n" + "END_DATA")

    return



from pathlib import Path

import click
import logging
from . import get_ctdcal_config

# Rich handling
# from rich.logging import RichHandler
# from rich.console import Console
# handler = logging.StreamHandler()
# handler.addFilter(logging.Filter("ctdcal"))  # filter out msgs from other modules
# FORMAT = "%(funcName)s: %(message)s"
# logging.basicConfig(
#     level="INFO",
#     format=FORMAT,
#     datefmt="[%X]",
#     handlers=[handler, RichHandler(console=Console(stderr=True))],
# )

# log = logging.getLogger(__name__)

handler = logging.StreamHandler()
handler.addFilter(logging.Filter("ctdcal"))  # filter out msgs from other modules
FORMAT = "%(funcName)s: %(message)s"
logging.basicConfig(
    level="NOTSET",
    format=FORMAT,
    datefmt="[%X]",
    handlers=[handler],
    # handlers=[RichHandler(console=Console(stderr=True))],
)

log = logging.getLogger(__name__)
cfg = get_ctdcal_config()


@click.group()
def cli():
    """The ctdcal command creates and manipulates data directories

    Documentation: tbd
    """
    pass


@cli.command()
def init():
    """Setup data folder with appropriate subfolders"""

    log.info(f"Building default /data/ directories: \n {*cfg.dirs.keys(),}")

    for sub_dir in cfg.dirs.values():
        Path(sub_dir).mkdir(parents=True)


@cli.command("import")  # click workaround to get a command named 'import'
def import_data():
    """Import data from given folder into ctdcal for processing"""
    # something like this?
    # ctdcal import _path_

    # TODO: smart imports based on file ext? .hex, .xmlcon, .cap
    # NOTE: ODF file types vs. others (oxygen, salt)

    pass


@cli.command()
@click.option(
    "-g",
    "--group",
    type=click.Choice(["ODF", "PMEL"], case_sensitive=False),
    default="ODF",
)
@click.option(
    "-t", "--type", type=click.Choice(["bottle", "ctd", "all"], case_sensitive=False)
)
def process(group, type):
    """Process data using a particular group's methodology"""

    if group == "ODF":
        from .scripts.odf_process_all import odf_process_all

        odf_process_all()
    elif group == "PMEL":
        # pmel_process()
        pass


@cli.command()
def cruise_report():
    """Generate bottle residual figures for cruise report"""

    from .scripts.cruise_report import cruise_report_residuals

    cruise_report_residuals()


if __name__ == "__main__":
    cli()



"""
This script will initialize the data directory for ctdcal according to ODF usage.
Joseph Gum, January 22, 2019
"""
import pathlib
import sys


def build_dirs():
    data_directory = "./data/"
    directories_to_be_made = [
        "raw",
        "salt",
        "bottle",
        "converted",
        "logs",
        "oxygen",
        "pressure",
        "quality_codes",
        "reft",
        "time",
    ]

    # will make parents directory if not missing, will not throw an error if directory already exists
    for x in directories_to_be_made:
        pathlib.Path(data_directory + x).mkdir(parents=True, exist_ok=True)


def main(argv):
    build_dirs()


if __name__ == "__main__":
    main(sys.argv[1:])



import xml.etree.ElementTree as ET


def sbe_xml_reader_a(file):
    """Function to read .XMLCON file from Seabird.

    Input:
    file: .XMLCON file produced from Seasave 7.

    Output:
    Dictionary of sensors.

    """
    tree = ET.parse(file)
    root = tree.getroot()

    """Pokedex is a dict of {Sensor index numbers from the config:sensor info}
    Assume that sensor index number is the order the sensors have been entered into the file.
    Therefore, it will be Frequency instruments first, then Voltage instruments.
    Call their index number (starting at 0) in order to pull out the info.

    """
    pokedex = {}
    for x in root.iter('Sensor'):
        """Start creating single sensor dictionary."""
        bulbasaur = {}
        for children in x:
            for y in children.iter():
                bulbasaur[y.tag] = float_convert(y.text)
            """Add sensor to big dictionary."""
            pokedex[x.attrib['index']] = bulbasaur
    return pokedex

###### Needs to be looked at ######

def number(string):
    """Helper function to convert strings to int or floats.
    Input: string
    Output: Int, Float or String if non convertable.
    """
    try:
        return int(string)
    except TypeError:
        return float(string)
    except ValueError:
        return string

"""Should be a way to do it with BAFP, need to try later."""
def number_convert(string):
    """Helper function to convert strings to int or floats.
    Input: string
    Output: Int, Float or String if non convertable.
    """
    try:
        return int(string)
    except ValueError:
        try:
            return float(string)
        except TypeError:
            return None
    except ValueError:
        return string

def number_convert_2(string):
    """Use regex to search determine if a string is an int, float, or other.
    Return as int, float or string respectively. Requires "import re".

    Input:
    string: The string to be converted to int, float or string.

    REGEX NOT FULLY TESTED

    """
    print(string)

    #regex patterns
    float_pattern = re.compile('-?[0-9]*\.?[0-9]+|-?[0-9]*\.?[0-9]+e[+-][0-9]*')
    int_pattern = re.compile('-?[0-9]+')

    try:
        if float_pattern.fullmatch(string) is True:
            return float(string)
        if int_pattern.fullmatch(string) is True:
            return int(string)
        return string
    #catch None
    except TypeError:
        return string

###### Needs to be looked at END ######

def float_convert(string):
    try:
        return float(string)
    except:
        return string



#!/bin/bash/env
"""
This module is used to parse the xml config and hex file provided by user. It
takes the xml config and extracts the useful settings used during a CTD cast,
and uses that to check and parse the hex file.
"""

import datetime
import re
import struct
import xml.etree.cElementTree as ET

import numpy as np
from pytz import timezone


class SBEReader:
    """Code originally written by Andrew Barna, January-March 2016."""

    def __init__(self, raw_hex, xml_config):
        """expects long character string inputs"""
        self.raw_hex = raw_hex
        self.xml_config = xml_config
        self._parse_config()
        self._load_hex()
        self._check_scan_lengths()

    def _load_hex(self):
        split_lines = self.raw_hex.splitlines()
        self.raw_bytes = [
            line.strip().encode("utf-8")
            for line in split_lines
            if not line.startswith("*")
        ]
        # next few lines are to grab start_scan_time
        self.raw_comments = [
            line.strip().split() for line in split_lines if line.startswith("*")
        ]

    def _check_scan_lengths(self):
        if not all([len(scan) == self.scan_length for scan in self.raw_bytes]):
            raise ValueError(
                "The data length does not match the expected length from the config"
            )

    def _parse_scans(self):
        """The order according to the SBE docs are:
        1) Data from the instrument
          a) Frequency (3 bytes each)
          b) Voltage (12 bits each)
        2) Surface Par (3 bytes)
        3) NMEA lat/lon (7 bytes)
        4) NMEA depth (3 bytes)
        5) NMEA time (4 bytes) (low byte first)
        6) Additional Data from the instrument
          a) Pressure temp (12 bits)
          b) pump status (4 bits)
          c) modulo byte (1 byte)
        7) System time (4 bytes) (low byte first)
        If any of the above are omitted, the length of the hex will be smaller.
        """

        # parse the frequencies
        # scan_list = list(scan)
        num_frequencies = 5 - self.config["FrequencyChannelsSuppressed"]
        num_voltages = 8 - self.config["VoltageWordsSuppressed"]
        flag_spar = 0

        if self.config["SurfaceParVoltageAdded"]:
            flag_spar = 1

        the_bytes = b"".join(self.raw_bytes)

        # specify how each line of raw data should be broken down
        unpack_str = (
            "6s" * num_frequencies
            + "3s" * num_voltages
            + "2s" * flag_spar
            + "4s" * flag_spar
            + "{}s".format(
                self.scan_length
                - num_voltages * 3
                - num_frequencies * 6
                - flag_spar * 6
            )
        )
        measurements = np.array(
            [
                [int(x, 16) for x in line]
                for line in struct.iter_unpack(unpack_str, the_bytes)
            ]
        )

        # this uses numpy magic to just "apply" the needed operations to the correct part of the array all in one go
        measurements[:, :num_frequencies] = measurements[:, :num_frequencies] / 256
        measurements[:, num_frequencies : num_frequencies + num_voltages] = 5 * (
            1
            - (measurements[:, num_frequencies : num_frequencies + num_voltages] / 4095)
        )
        measurements = measurements[
            :, 0:-1
        ]  # remove after applying _parse_scans_meta in here
        # print(measurements.shape)
        return measurements

    def _parse_scans_meta(self):
        """The order according to the SBE docs are:
        1) Data from the instrument
          a) Frequency (3 bytes each)
          b) Voltage (12 bits each)
        2) Surface Par (3 bytes)
        3) NMEA lat/lon (7 bytes)
        4) NMEA depth (3 bytes)
        5) NMEA time (4 bytes) (low byte first)
        6) Additional Data from the instrument
          a) Pressure temp (12 bits)
          b) pump status (4 bits)
          c) modulo byte (1 byte)
        7) System time (4 bytes) (low byte first)
        If any of the above are omitted, the length of the hex will be smaller.
        """

        # parse the frequencies
        # scan_list = list(scan)
        num_frequencies = 5 - self.config["FrequencyChannelsSuppressed"]
        num_voltages = 8 - self.config["VoltageWordsSuppressed"]

        string_order = ["scan"]
        # flags to determine how many bytes to break by
        flag_nmea_pos = 0
        flag_nmea_depth = 0
        flag_nmea_time = 0
        flag_scan_time = 0
        flag_spar = 0
        # put in to make code readable later on
        flag_pressure_temp = 1
        flag_ctd_status = 1
        flag_modulo = 1
        if self.config["SurfaceParVoltageAdded"]:
            flag_spar = 1
            # string_order.append('spar') #already added into mass line, does not need additional flag
        if self.config["NmeaPositionDataAdded"]:
            flag_nmea_pos = 1
            string_order.append("nmea_pos")
            # Depth is here for completeness but not implemented,
            # after email chain showed SBE no longer knows how they did it.
        if self.config["NmeaDepthDataAdded"]:
            flag_nmea_depth = 1
            string_order.append("nmea_depth")
        if self.config["NmeaTimeAdded"]:
            flag_nmea_time = 1
            string_order.append("nmea_time")
        string_order.append("pressure_temp")
        string_order.append("flag_ctd_status")
        string_order.append("modulo")
        if self.config["ScanTimeAdded"]:
            flag_scan_time = 1
            string_order.append("scan_time")

        the_bytes = b"".join(self.raw_bytes)

        # specify how each line of raw data should be broken down
        # this format is fed directly into _breakdown, so it needs to be tracked there
        unpack_str = (
            str(num_frequencies * 6 + num_voltages * 3 + flag_spar * 6)
            + "s"
            + "14s" * flag_nmea_pos
            + "6s" * flag_nmea_depth
            + "8s" * flag_nmea_time
            + "3s" * flag_pressure_temp
            + "1s" * flag_ctd_status
            + "2s" * flag_modulo
            + "8s" * flag_scan_time
        )
        # breaks down line according to columns specified by unpack_str and converts from hex to decimal MSB first
        # needs to be adjusted for LSB fields (NMEA time, scan time), break into two lines? move int(x,16) outside
        measurements = [line for line in struct.iter_unpack(unpack_str, the_bytes)]
        # print(unpack_str)
        # measurements_2 = np.array([self._breakdown(line, string_order) for line in measurements])
        measurements_2 = [self._breakdown(line, string_order) for line in measurements]
        # if no time is enabled, fake the scan timestamp from info in the .hex file
        if flag_scan_time == 0:
            measurements_3 = self._sbe_time_seq(measurements_2)
            #            print(True)
            measurements_3 = np.array(measurements_3)
            return measurements_3
        # print(measurements_3)
        return np.array(measurements_2)

    def _breakdown(self, line, ordering):
        """
        Convert hex metadata to science useable data.

        Steps:
        1. Throw away away first part of the line
        2. Start processing the rest of the line
        3. Return a sequence (tuple? list?)

        Input:
        line - a sequence specified by format in _parse_scans_meta
        ordering - the order the string comes in

        Output:
        output - a string of converted data according to format

        Format:
        Lat, Lon, pressure temp, bottle fire status, NMEA time, Scan time
        """
        output = ""
        # print(ordering, line)
        for x, y in zip(ordering, line):
            if x == "nmea_pos":
                tokens = []
                # print(y)
                for t in struct.iter_unpack("2s2s2s2s2s2s2s", y):
                    for tt in t:
                        tokens.append(tt)
                output = (
                    output
                    + str(
                        self._location_fix(
                            tokens[0],
                            tokens[1],
                            tokens[2],
                            tokens[3],
                            tokens[4],
                            tokens[5],
                            tokens[6],
                        )
                    )
                    + ","
                )
            elif x == "nmea_time":
                output = (
                    output + str(self._sbe_time(self._reverse_bytes(y), "nmea")) + ","
                )
            elif x == "scan_time":
                output = output + str(self._sbe_time(self._reverse_bytes(y), "scan"))
            elif x == "flag_ctd_status":
                output = (
                    output
                    + str(self._pump_status(y))
                    + ","
                    + str(self._bottle_fire(y))
                    + ","
                )  # need to make final comma dynamic later so less errors down the line
            elif x == "pressure_temp":
                output = output + str(int(y, 16)) + ","

        return output

    def _breakdown_header(self):
        """Creates header for metadata. Arrays below are what is expected.

        ['GPSLAT', 'GPSLON', 'new_fix', 'nmea_datetime', 'pressure_temp_int', 'btl_fire', 'scan_datetime'],
        ['float64', 'float64', 'bool_', 'datetime64', 'int_', 'bool_', 'datetime64']
        """
        # temp fix, need to adjust to take in file to adjust as wanted?
        output = [[], []]
        if self.config["NmeaPositionDataAdded"]:
            output[0] = ["GPSLAT", "GPSLON", "new_fix"]
            output[1] = ["float64", "float64", "bool_"]
        if self.config["NmeaTimeAdded"]:
            output[0].append("nmea_datetime")
            output[1].append("float64")
        output[0].append("pressure_temp_int")
        output[1].append("int_")
        output[0].append("pump_on")
        output[1].append("bool_")
        output[0].append("btl_fire")
        output[1].append("bool_")
        # if self.config["ScanTimeAdded"]: #need to fix flagging part
        # output[0].append('scan_datetime')
        # output[1].append('float64')
        output[0].append("scan_datetime")
        output[1].append("float64")

        return output

    def _location_fix(self, b1, b2, b3, b4, b5, b6, b7):
        """Determine location from SBE format.

        Input:
        b1, b2, b3: three components of Latitude
        b4, b5, b6: three components of Longitude
        b7: sign for lat/lon, is it a new fix

        Output:
        Tuple with three components in order:
        latitude, longitude, new fix
        Latitude is a float
        Longitude is a float
        If it is a new fix it will be true, otherwise false
        """
        lat = (int(b1, 16) * 65536 + int(b2, 16) * 256 + int(b3, 16)) / 50000
        lon = (int(b4, 16) * 65536 + int(b5, 16) * 256 + int(b6, 16)) / 50000

        """
        If bit 1 in byte_pos is 1, this is a new position
        If bit 8 in byte_pos is 1, lat is negative
        If bit 7 in byte_pos is 1, lon is negative
        """
        flag_new_fix = False

        mask_lat_pos = 0x80
        mask_lon_pos = 0x40
        mask_new_fix = 0x01

        if int(b7, 16) & mask_lat_pos:
            lat = lat * -1
        if int(b7, 16) & mask_lon_pos:
            lon = lon * -1
        if int(b7, 16) & mask_new_fix:
            flag_new_fix = True

        output = "{0},{1},{2}".format(lat, lon, flag_new_fix)
        return output

    def _reverse_bytes(self, hex_time):
        """Reverse hex time according to SBE docs.
        Split number by every two chars, then recombine them in reverse order.

        This makes no assumptions about length or whatever. If there are an odd number of characters
        it adds a 0 before the string to pad it out.

        The final output is guaranteed to have '0x' at the beginning.
        """
        time = hex_time.decode("utf-8")
        if re.match("0x", time):
            time = time[2:]
        if (len(time) % 2) == 1:
            time = "0" + time

        """Start flipping."""
        list_1 = []
        reverse_hex = ""
        for y in range(0, len(time), 2):
            list_1.append(time[y : y + 2])
        for y in range(len(list_1) - 1, -1, -1):
            reverse_hex = reverse_hex + list_1[y]
        """Add prefix to make python hex compatible"""
        reverse_hex = "0x" + reverse_hex
        # print('Output of reverse_hex: ' + reverse_hex)
        return reverse_hex

    def _sbe_time(self, hex_time, sbe_type):
        """Convert raw data from sbe .hex file to appropriate datetime, then return string.
        Assumes hex_string has already been reversed by every two chars.
        Needs to be checked for timezone problems in the future.

        Input:
        hex_time: raw hex corresponding to nmea time or scan time. Must be 8 characters.
        sbe_type: either "nmea" or "scan"

        Output:
        A string to be ready printed to csv.

        """
        # values as state by SBE in manual
        scan_start = datetime.datetime(1970, 1, 1, 0, 0, 0)
        nmea_start = datetime.datetime(2000, 1, 1, 0, 0, 0)

        # catch if 8 characters was not input
        if re.match("0x", hex_time):
            hex_time = hex_time[2:]
        if len(hex_time) > 8:
            raise Exception("Hex string too long to be SBE formatted time: " + hex_time)
        if len(hex_time) < 8:
            raise Exception(
                "Hex string too short to be SBE formatted time: " + hex_time
            )

        seconds = datetime.timedelta(seconds=int(hex_time, 16))
        # changed to epoch time at request of analysts
        # epoch time assumes it is UTC all way through
        if sbe_type == "scan":
            # time = scan_start + seconds
            # return time.isoformat()
            time = scan_start + seconds
            return time.replace(tzinfo=timezone("UTC")).timestamp()
        elif sbe_type == "nmea":
            # time = nmea_start + seconds
            # return time.isoformat()
            time = nmea_start + seconds
            return time.replace(tzinfo=timezone("UTC")).timestamp()
        else:
            raise Exception(
                'Please choose "nmea" or "scan" for second input to _sbe_time()'
            )

    def _sbe_time_create(self, utc_time, sbe_type="scan"):
        """Reverse of _sbe_time, create a sbe timestamp in hex to append to file.
        Useful when SBE acquisition software does not have NMEA time or scan time activated.

        The inverse of _sbe_time

        Input:
        utc_time: datetime
        """

        # values as state by SBE in manual
        scan_start = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone("UTC"))
        nmea_start = datetime.datetime(2000, 1, 1, 0, 0, 0, tzinfo=timezone("UTC"))
        utc_time = utc_time.replace(tzinfo=timezone("UTC"))
        if sbe_type == "scan":
            time = utc_time - scan_start
        elif sbe_type == "nmea":
            time = utc_time - nmea_start

        # timedelta method to return seconds from any timedelta format
        hex_time = hex(int(time.total_seconds()))

        output = self._reverse_bytes(bytearray(hex_time, "utf-8"))
        return output

    def _sbe_time_seq(self, data_list):
        """Recreates the scan timestamp if the option was not enabled in SBE acq.
        Accurate to 1 second/24hz, as it uses the start time in the second to last line of the .hex file.

        Sequencer will run over a processed file/list? and append the scan time at the end.
        """
        # Pull out the second to last line of the comments in .hex,
        # then pull out the datetime info at the end of the line, then format
        start_month = self.raw_comments[-2][-4]
        start_day = self.raw_comments[-2][-3]
        start_year = self.raw_comments[-2][-2]
        start_time = self.raw_comments[-2][-1]

        start_scan_time = datetime.datetime.strptime(
            start_month + start_day + start_year + start_time, "%b%d%Y%H:%M:%S"
        )
        current_scan_time = start_scan_time
        hz_counter = 0
        output = []
        for line in data_list:
            if hz_counter >= 24:
                hz_counter = 0
                current_scan_time = current_scan_time + datetime.timedelta(seconds=1)
            line = line + str(
                current_scan_time.replace(tzinfo=timezone("UTC")).timestamp()
            )
            output.append(line)
            hz_counter += 1
        return output

    def _flag_status(self, flag_char, scan_number):
        """Decode SBE flag bit, as referenced on SBE 11pV2, pg 66.
        CTD status:
            Bit 0 Pump status- 1=pump on, 0=pump off.
            Bit 1 Bottom contact switch status -
                1 = switch open (no contact), 0 = switch closed.
            Bit 2 G.O. 1015 water sampler interface confirm signal or
                manual pump control signal -
                1 = Deck Unit detects confirm signal from G.O. 1015 or detects manual pump control installed in 9plus,
                0 = not detected.
            Bit 3 CTD modem carrier detect -
                0 = CTD modem detects Deck Unit modem carrier signal, 1 = not detected.

        Output could be cleaned up if it ever needs to be used.
        """

        mask_pump = 0x1
        mask_bottom_contact = 0x2
        mask_water_sampler_interface = 0x4
        mask_modem_carrier_detect = 0x8

        output = ""

        if flag_char & mask_pump:
            output += "pump on, "
        else:
            output += "pump off, "
        if flag_char & mask_bottom_contact:
            output += "no bottom contact, "
        else:
            output += "bottom contact, "
        if flag_char & mask_water_sampler_interface:
            output += (
                "G.O. 1015 water sampler confirm signal OR manual pump installed, "
            )
        else:
            output += "no signal from G.O. 1015 water sampler OR no manual pump, "
        if flag_char & mask_modem_carrier_detect:
            output += "modem carrier signal detected"
        else:
            output += "no modem carrier signal detected"

        return output

    def _pump_status(self, flag_char):
        """Decode SBE flag bit, as referenced on SBE 11pV2, pg 66.
        CTD status:
            Bit 0 Pump status- 1=pump on, 0=pump off.
        """
        mask_pump = 1
        if bytes.decode(flag_char).isnumeric() is False:
            flag_char = b"3"
        if int(flag_char) & mask_pump:
            return True
        else:
            return False

    def _bottle_fire(self, flag_char):
        """Determine if a scan is around a bottle firing as marked by SBE.
        A simplification of _flag_status.

        bit 4 = bottle fire status, 1 = fired, 0 = not fired
        bit 4 when set notes a bottle has been fired, no positional info

        Input:
        flag_char - a single hex char that

        Output:
        Returns True if scan is a part of a bottle fire,
        or False if scan is not a part of a bottle fire.

        Need to check for off-by-one errors
        """

        mask_bottle_fire = 0x4

        if int(flag_char, 16) & mask_bottle_fire:
            return True
        else:
            return False

        return None

    # no calls to this, what are AD590M and AD590B? calib coefs prob
    # def _digiquartz_temp_correction(self, hex_in):
    #     """Digiquartz pressure sensor temperature correction for internal probe."""
    #     charm = format(hex_in, "d")
    #     T_d = (AD590M * charm) + AD590B
    #     return T_d

    """Only for values in bools and numeric. """

    def _parse_config(self):
        bools = [
            "SurfaceParVoltageAdded",
            "NmeaPositionDataAdded",
            "NmeaDepthDataAdded",
            "NmeaTimeAdded",
            "ScanTimeAdded",
        ]
        numeric = [
            "FrequencyChannelsSuppressed",
            "VoltageWordsSuppressed",
        ]
        sensors = {}

        config = ET.fromstring(self.xml_config)
        self.config = {}

        for key in numeric:
            try:
                self.config[key] = int(config.find("./Instrument/{}".format(key)).text)
            except AttributeError:
                raise AttributeError("Could not find {} in XMLCONF".format(key))
            except ValueError:
                raise ValueError("{} Value is not a number".format(key))

        for key in bools:
            try:
                self.config[key] = bool(
                    int(config.find("./Instrument/{}".format(key)).text)
                )
            except AttributeError:
                raise AttributeError("Could not find {} in XMLCONF".format(key))
            except ValueError:
                raise ValueError("{} Value is not truthy".format(key))

        """Pokedex is a dict of {Sensor index numbers from the config:sensor info}
        Assume that sensor index number is the order the sensors have been entered into the file.
        Therefore, it will be Frequency instruments first, then Voltage instruments.
        Call their index number (starting at 0) in order to pull out the info.

        """
        # pokedex = {}
        for x in config.iter("Sensor"):
            # print(ET.tostring(x))
            """Start creating single sensor dictionary."""
            bulbasaur = {}
            bulbasaur["SensorID"] = x.attrib["SensorID"]
            # load all values into dict - beware of overwriting #NEED TO FIX
            for children in x:
                for y in children.iter():
                    try:
                        bulbasaur[y.tag] = float(y.text)
                    except (TypeError, ValueError):
                        bulbasaur[y.tag] = (
                            str(y.text).replace("\n", "").replace(" ", "")
                        )

                """Add sensor to big dictionary."""
                # pokedex[x.attrib['index']] = bulbasaur
                sensors[int(x.attrib["index"])] = bulbasaur
        # sensors.append(pokedex)
        self.config["Sensors"] = sensors

    @classmethod
    def from_paths(cls, raw_hex_path, xml_config_path, encoding="cp437"):
        with open(raw_hex_path, encoding=encoding) as raw_hex_file, open(
            xml_config_path, encoding=encoding
        ) as xml_config_file:
            return cls(raw_hex_file.read(), xml_config_file.read())

    @property
    def scan_length(self):
        return self.calculate_scan_byte_length()

    @property
    def frequency_channels(self):
        pass

    def voltage_channels(self):
        pass

    def calculate_scan_byte_length(self):
        length = 6  # Extra info alway present takes 6 "chars"
        length += 6 * (5 - self.config["FrequencyChannelsSuppressed"])
        length += 3 * (8 - self.config["VoltageWordsSuppressed"])
        length += 8 * self.config["ScanTimeAdded"]
        length += 8 * self.config["NmeaTimeAdded"]
        length += 6 * self.config["NmeaDepthDataAdded"]
        length += 14 * self.config["NmeaPositionDataAdded"]
        length += 6 * self.config["SurfaceParVoltageAdded"]
        return int(length)

    @property
    def parsed_scans(self):
        """
        Wrapper for _parse_scans and _parse_scans_meta. Returns np.ndarray
        """
        # try:
        #     return self._parse_scans
        # except AttributeError:
        #     self._parse_scans = np.concatenate((self._parse_scans(), self._parse_scans_meta().reshape(self._parse_scans_meta().size,1)), axis = 1)
        #     return self._parse_scans
        return np.concatenate(
            (
                self._parse_scans(),
                self._parse_scans_meta().reshape(self._parse_scans_meta().size, 1),
            ),
            axis=1,
        )
        # return self._parse_scans

    def parsed_config(self):
        return self.config

    def to_dict(self, parse_cache=True):
        return {
            "raw_hex": self.raw_hex,
            "xml_config": self.xml_config,
            "_parsed_scans": self.parsed_scans,
        }

    @classmethod
    def from_dict(cls, data):
        instance = cls(data["raw_hex"], data["xml_config"])
        try:
            instance._parsed_scans = data["_parsed_scans"]
        except KeyError:
            pass
        return instance



#!/usr/bin/env python
from pathlib import Path

import gsw
import logging
import numpy as np
import pandas as pd
from scipy.ndimage.interpolation import shift
import yaml

from . import convert as convert
from . import ctd_plots as ctd_plots
from . import flagging as flagging
from . import get_ctdcal_config
from . import process_ctd as process_ctd

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)


def load_fit_yaml(fname=f"{cfg.dirs['logs']}fit_coefs.yaml", to_object=False):
    """Load polynomail fit order information from .yaml file."""

    with open(fname, "r") as f:
        ymlfile = yaml.safe_load(f)

    if to_object:
        return type("ymlfile", (object,), ymlfile)
    else:
        return ymlfile


def write_fit_yaml():
    """For future use with automated fitting routine(s).
    i.e., iterate to find best fit parameters, save to file"""
    pass


def cell_therm_mass_corr(temp, cond, sample_int=1 / 24, alpha=0.03, beta=1 / 7):
    """Correct conductivity signal for effects of cell thermal mass.

    Parameters
    ----------
    temp : array-like
        CTD temperature [degC]
    cond : array-like
        CTD conductivity [mS/cm]
    sample_int : float, optional
        CTD sample interval [seconds]
    alpha : float, optional
        Thermal anomaly amplitude
    beta : float, optional
        Thermal anomaly time constant

    Returns
    -------
    cond_corr : array-like
        Corrected CTD conductivity [mS/cm]

    Notes
    -----
    See Sea-Bird Seasoft V2 manual (Section 6, page 93) for equation information.
    Default alpha/beta values taken from Seasoft manual (page 92).
    c.f. "Thermal Inertia of Conductivity Cells: Theory" (Lueck 1990) for more info
    https://doi.org/10.1175/1520-0426(1990)007<0741:TIOCCT>2.0.CO;2
    """
    a = 2 * alpha / (sample_int * beta + 2)
    b = 1 - (2 * a / alpha)
    dc_dT = 0.1 * (1 + 0.006 * (temp - 20))
    dT = np.insert(np.diff(temp), 0, 0)  # forward diff reduces len by 1

    def calculate_CTM(b, CTM_0, a, dc_dT, dT):
        """Return CTM in units of [S/m]"""
        CTM = -1.0 * b * CTM_0 + a * (dc_dT) * dT
        return CTM

    CTM = calculate_CTM(b, 0, a, dc_dT, dT)
    CTM = calculate_CTM(b, shift(CTM, 1, order=0), a, dc_dT, dT)
    CTM = np.nan_to_num(CTM) * 10.0  # [S/m] to [mS/cm]
    cond_corr = cond + CTM

    return cond_corr


def _flag_btl_data(
    df,
    param=None,
    ref=None,
    thresh=[0.002, 0.005, 0.010, 0.020],
    f_out=None,
):
    """
    Flag CTD "btl" data against reference measurement (e.g. SBE35, bottle salts).

    Parameters
    ----------
    df : DataFrame,
        DataFrame containing btl data
    param : str
        Name of parameter to calibrate (e.g. "CTDCOND1", "CTDTMP2")
    ref : str
        Name of reference parameter to calibrate against (e.g. "BTLCOND", "T90")
    thresh : list of float, optional
        Maximum acceptable residual for each pressure range
    f_out : str, optional
        Path and filename to save residual vs. pressure plots

    Returns
    -------
    df_ques : DataFrame
        Data flagged as questionable (flag 3s)
    df_bad : DataFrame
        Data flagged as bad (flag 4s)

    """
    # TODO: thresh should probably be put in config/cast-by-cast config
    prs = cfg.column["p"]

    # Remove extreme outliers and code bad
    df = df.reset_index(drop=True)
    df["Diff"] = df[ref] - df[param]
    df["Flag"] = flagging.outliers(df["Diff"])

    # Find values that are above the threshold and code questionable
    df["Flag"] = flagging.by_residual(df[param], df[ref], df[prs], old_flags=df["Flag"])
    df_good = df[df["Flag"] == 2].copy()
    df_ques = df[df["Flag"] == 3].copy()
    df_bad = df[df["Flag"] == 4].copy()

    if f_out is not None:
        if param == cfg.column["t1"]:
            xlabel = "T1 Residual (T90 C)"
        elif param == cfg.column["t2"]:
            xlabel = "T2 Residual (T90 C)"
        elif param == cfg.column["c1"]:
            xlabel = "C1 Residual (mS/cm)"
        elif param == cfg.column["c2"]:
            xlabel = "C2 Residual (mS/cm)"
        f_out = f_out.split(".pdf")[0] + "_postfit.pdf"
        ctd_plots._intermediate_residual_plot(
            df["Diff"],
            df[prs],
            df["SSSCC"],
            show_thresh=True,
            xlabel=xlabel,
            f_out=f_out,
        )
        f_out = f_out.split(".pdf")[0] + "_flag2.pdf"
        ctd_plots._intermediate_residual_plot(
            df_good["Diff"],
            df_good[prs],
            df_good["SSSCC"],
            show_thresh=True,
            xlabel=xlabel,
            f_out=f_out,
        )

    return df_ques, df_bad


def _prepare_fit_data(df, param, ref_param, zRange=None):
    """Remove non-finite data, trim to desired zRange, and remove extreme outliers"""

    good_data = df[np.isfinite(df[ref_param]) & np.isfinite(df[param])].copy()

    if zRange is not None:
        zMin, zMax = zRange.split(":")
        good_data = good_data[
            (good_data["CTDPRS"] > int(zMin)) & (good_data["CTDPRS"] < int(zMax))
        ]

    good_data["Diff"] = good_data[ref_param] - good_data[param]
    good_data["Flag"] = flagging.outliers(good_data["Diff"], n_sigma2=4)

    df_good = good_data[good_data["Flag"] == 2].copy()
    df_bad = good_data[good_data["Flag"] == 4].copy()

    return df_good, df_bad


def multivariate_fit(y, *args, coef_names=None, const_name="c0"):
    """
    Least-squares fit data using multiple dependent variables. Dependent variables must
    be provided in tuple pairs of (data, order) as positional arguments.

    If coef_names are defined, coefficients will be returned as a dict. Otherwise,
    coefficients are return as an array in the order of the dependent variables, sorted
    by decreasing powers.

    Parameters
    ----------
    y : array-like
        Indepedent variable to be fit
    args : tuple
        Pairs of dependent variable data and fit order (i.e., (data, order))
    coef_names : list-like, optional
        Base names for coefficients (i.e., "a" for 2nd order yields ["a2", "a1"])
    const_name : str, optional
        Name for constant offset term

    Returns
    -------
    coefs : array-like
        Least-squares fit coefficients in decreasing powers

    Examples
    --------
    Behavior when coef_names is None:

    >>> z = [1, 4, 9]
    >>> x = [1, 3, 5]
    >>> y = [1, 2, 3]
    >>> multivariate_fit(z, (x, 2), (y, 1))
    array([0.25, 0.375, 0.25, 0.125])  # [c1, c2, c3, c4]

    where z = (c1 * x ** 2) + (c2 * x) + (c3 * y) + c4

    Behavior when coef_names is given:

    >>> z = [1, 4, 9]
    >>> x = [1, 3, 5]
    >>> y = [1, 2, 3]
    >>> multivariate_fit(z, (x, 2), (y, 1), coef_names=["a", "b"])
    {"a2": 0.25, "a1": 0.375, "b1": 0.25, "c0": 0.125}

    where z = (a2 * x ** 2) + (a1 * x) + (b1 * y) + c0
    """
    to_dict = True
    if coef_names is None:
        to_dict = False
        coef_names = [""] * len(args)  # needed to prevent zip() error
    elif len(args) != len(coef_names):
        raise ValueError(
            "length of coef_names must match the number of dependent variables"
        )

    # iteratively build fit matrix
    rows, names = [], []
    for arg, coef_root in zip(args, coef_names):
        if type(arg) is not tuple:
            raise TypeError(f"Positional args must be tuples, not {type(arg)}")

        series, order = arg
        for n in np.arange(1, order + 1)[::-1]:
            rows.append(series ** n)  # n is np.int64 so series will cast to np.ndarray
            names.append(f"{coef_root}{n}")

    # add constant offset term
    rows.append(np.ones(len(y)))
    names.append(const_name)

    fit_matrix = np.vstack(rows)
    coefs = np.linalg.lstsq(fit_matrix.T, y, rcond=None)[0]

    return dict(zip(names, coefs)) if to_dict else coefs


def apply_polyfit(y, y_coefs, *args):
    """
    Apply a polynomial correction to series of data. Coefficients should be provided in
    increasing order (i.e., a0, a1, a2 for y_fit = y + a2 * y ** 2 + a1 * y + a0)

    For the independent variables (y), coefficients start from the zero-th order (i.e.,
    constant offset). For dependent variables (args), coefficients start from the first
    order (i.e., linear term).

    Parameters
    ----------
    y : array-like
        Independent variable data to be corrected
    y_coefs : tuple of float
        Independent variable fit coefficients (i.e., (data, coef0, ..., coefN))
    args : tuple of (array-like, (float, float, ...))
        Dependent variable data and fit coefficients (i.e., (data, coef1, ..., coefN))

    Returns
    -------
    fitted_y : array-like
        Independent variable data with polynomial fit correction applied
    """
    fitted_y = np.copy(y)
    for n, coef in enumerate(y_coefs):
        fitted_y += coef * y ** n

    for arg in args:
        if type(arg) is not tuple:
            raise TypeError(f"Positional args must be tuples, not {type(arg)}")

        series, coefs = arg
        for n, coef in enumerate(coefs):
            fitted_y += coef * series ** (n + 1)

    return fitted_y


def calibrate_temp(btl_df, time_df):
    # TODO: make this an ODF script in ctdcal/scripts?
    # TODO: break off parts of this to useful functions for all vars (C/T/O)
    """
    Least-squares fit CTD temperature data against reference data.

    Parameters
    -----------
    btl_df : DataFrame
        CTD data at bottle stops
    time_df : DataFrame
        Continuous CTD data

    Returns
    --------

    """
    log.info("Calibrating temperature")
    ssscc_subsets = sorted(Path(cfg.dirs["ssscc"]).glob("ssscc_t*.csv"))
    if not ssscc_subsets:  # if no t-segments exists, write one from full list
        log.debug(
            "No CTDTMP grouping file found... creating ssscc_t1.csv with all casts"
        )
        if not Path(cfg.dirs["ssscc"]).exists():
            Path(cfg.dirs["ssscc"]).mkdir()
        ssscc_list = process_ctd.get_ssscc_list()
        ssscc_subsets = [Path(cfg.dirs["ssscc"] + "ssscc_t1.csv")]
        pd.Series(ssscc_list).to_csv(ssscc_subsets[0], header=None, index=False)

    fit_yaml = load_fit_yaml()  # load fit polynomial order
    for tN in ["t1", "t2"]:
        T_flag, T_fit_coefs = pd.DataFrame(), pd.DataFrame()
        for f in ssscc_subsets:
            # 0) load ssscc subset to be fit together
            ssscc_sublist = pd.read_csv(f, header=None, dtype="str").squeeze().to_list()
            btl_rows = btl_df["SSSCC"].isin(ssscc_sublist).values
            good_rows = btl_rows & (btl_df["REFTMP_FLAG_W"] == 2)
            time_rows = time_df["SSSCC"].isin(ssscc_sublist).values

            # 1) plot pre-fit residual
            f_stem = f.stem  # get "ssscc_t*" from path
            ctd_plots._intermediate_residual_plot(
                btl_df.loc[btl_rows, cfg.column["refT"]]
                - btl_df.loc[btl_rows, cfg.column[tN]],
                btl_df.loc[btl_rows, cfg.column["p"]],
                btl_df.loc[btl_rows, "SSSCC"],
                xlabel=f"{tN.upper()} Residual (T90 C)",
                f_out=f"{cfg.fig_dirs[tN]}residual_{f_stem}_prefit.pdf",
            )

            # 2) prepare data for fitting
            # NOTE: df_bad will be overwritten during post-fit data flagging but is
            # left here for future debugging (if necessary)
            df_good, df_bad = _prepare_fit_data(
                btl_df[good_rows],
                cfg.column[tN],
                cfg.column["refT"],
                zRange=fit_yaml[tN][f_stem]["zRange"],
            )
            ctd_plots._intermediate_residual_plot(
                df_good["Diff"],
                df_good[cfg.column["p"]],
                df_good["SSSCC"],
                xlabel=f"{tN.upper()} Residual (T90 C)",
                f_out=f"{cfg.fig_dirs[tN]}residual_{f_stem}_fit_data.pdf",
            )

            # 3) calculate fit coefs
            # TODO: truncate coefs (10 digits? look at historical data)
            P_order = fit_yaml[tN][f_stem]["P_order"]
            T_order = fit_yaml[tN][f_stem]["T_order"]
            coef_dict = multivariate_fit(
                df_good["Diff"],
                (df_good[cfg.column["p"]], P_order),
                (df_good[cfg.column[tN]], T_order),
                coef_names=["cp", "ct"],
            )

            # 4) apply fit
            P_coefs = tuple(coef_dict[f"cp{n}"] for n in np.arange(1, P_order + 1))
            T_coefs = tuple(coef_dict[f"ct{n}"] for n in np.arange(1, T_order + 1))
            btl_df.loc[btl_rows, cfg.column[tN]] = apply_polyfit(
                btl_df.loc[btl_rows, cfg.column[tN]],
                (coef_dict["c0"],) + T_coefs,
                (btl_df.loc[btl_rows, cfg.column["p"]], P_coefs),
            )
            time_df.loc[time_rows, cfg.column[tN]] = apply_polyfit(
                time_df.loc[time_rows, cfg.column[tN]],
                (coef_dict["c0"],) + T_coefs,
                (time_df.loc[time_rows, cfg.column["p"]], P_coefs),
            )

            # 4.5) flag CTDTMP and make residual plots
            df_ques, df_bad = _flag_btl_data(
                btl_df[btl_rows],
                param=cfg.column[tN],
                ref=cfg.column["refT"],
                f_out=f"{cfg.fig_dirs[tN]}residual_{f_stem}.pdf",
            )

            # 5) handle quality flags
            T_flag = pd.concat([T_flag, df_bad, df_ques])

            # 6) handle fit params
            coef_df = pd.DataFrame()
            coef_df["SSSCC"] = ssscc_sublist
            coef_names = ["cp2", "cp1", "ct2", "ct1", "c0"]
            coef_df[coef_names] = 0.0
            for k, v in coef_dict.items():
                coef_df[k] = v

            T_fit_coefs = pd.concat([T_fit_coefs, coef_df])

        # one more fig with all cuts
        ctd_plots._intermediate_residual_plot(
            btl_df[cfg.column["refT"]] - btl_df[cfg.column[tN]],
            btl_df[cfg.column["p"]],
            btl_df["SSSCC"],
            xlabel=f"{tN.upper()} Residual (T90 C)",
            show_thresh=True,
            f_out=f"{cfg.fig_dirs[tN]}residual_all_postfit.pdf",
        )

        # export temp quality flags
        # TODO: make these flags useful/less cluttered
        T_flag.sort_index().to_csv(f"{cfg.dirs['logs']}qual_flag_{tN}.csv", index=False)

        # export temp fit params (formated to 5 sig figs, scientific notation)
        T_fit_coefs[coef_names] = T_fit_coefs[coef_names].applymap(
            lambda x: np.format_float_scientific(x, precision=4, exp_digits=1)
        )
        T_fit_coefs.to_csv(cfg.dirs["logs"] + f"fit_coef_{tN}.csv", index=False)

    # flag temperature data
    # TODO: CTDTMP_FLAG_W historically not included in hy1 file... should it be?
    time_df["CTDTMP_FLAG_W"] = 2  # TODO: flag w/ REFT somehow? discrete vs continuous

    return True


def calibrate_cond(btl_df, time_df):
    # TODO: make this an ODF script in ctdcal/scripts?
    # TODO: break off parts of this to useful functions for all vars (C/T/O)
    # TODO: salt subset lists aren't loading in increasing order:
    # (still functions properly but the fit_coef_c#.csv is confusing as a result)
    """
    Least-squares fit CTD conductivity data against bottle salts.

    Parameters
    -----------
    btl_df : DataFrame
        CTD data at bottle stops
    time_df : DataFrame
        Continuous CTD data

    Returns
    --------

    """
    log.info("Calibrating conductivity")
    # calculate BTLCOND values from autosal data
    btl_df[cfg.column["refC"]] = convert.CR_to_cond(
        btl_df["CRavg"],
        btl_df["BathTEMP"],
        btl_df[cfg.column["t1"]],
        btl_df[cfg.column["p"]],
    )

    # merge in handcoded salt flags
    # TODO: make salt flagger move .csv somewhere else? or just always have it
    # somewhere else and read it from that location (e.g. in data/scratch_folder/salts)
    salt_file = "tools/salt_flags_handcoded.csv"  # abstract to config.py
    if Path(salt_file).exists():
        handcoded_salts = pd.read_csv(
            salt_file, dtype={"SSSCC": str, "salinity_flag": int}
        )
        handcoded_salts = handcoded_salts.rename(
            columns={"SAMPNO": "btl_fire_num", "salinity_flag": "SALNTY_FLAG_W"}
        ).drop(columns=["diff", "Comments"])
        btl_df = btl_df.merge(handcoded_salts, on=["SSSCC", "btl_fire_num"], how="left")
        # TODO: may be easier to try using flagging._merge_flags()?
        btl_df["SALNTY_FLAG_W"] = flagging.nan_values(
            btl_df["SALNTY"], old_flags=btl_df["SALNTY_FLAG_W"]
        )
    else:
        btl_df["SALNTY_FLAG_W"] = flagging.nan_values(btl_df["SALNTY"])

    ssscc_subsets = sorted(Path(cfg.dirs["ssscc"]).glob("ssscc_c*.csv"))
    if not ssscc_subsets:  # if no c-segments exists, write one from full list
        log.debug(
            "No CTDCOND grouping file found... creating ssscc_c1.csv with all casts"
        )
        ssscc_list = process_ctd.get_ssscc_list()
        ssscc_subsets = [Path(cfg.dirs["ssscc"] + "ssscc_c1.csv")]
        pd.Series(ssscc_list).to_csv(ssscc_subsets[0], header=None, index=False)

    fit_yaml = load_fit_yaml()  # load fit polynomial order
    for cN, tN in zip(["c1", "c2"], ["t1", "t2"]):
        C_flag, C_fit_coefs = pd.DataFrame(), pd.DataFrame()
        for f in ssscc_subsets:
            # 0) grab ssscc chunk to fit
            ssscc_sublist = pd.read_csv(f, header=None, dtype="str").squeeze().to_list()
            btl_rows = btl_df["SSSCC"].isin(ssscc_sublist).values
            good_rows = btl_rows & (btl_df["SALNTY_FLAG_W"] == 2)
            time_rows = time_df["SSSCC"].isin(ssscc_sublist).values

            # 1) plot pre-fit residual
            f_stem = f.stem  # get "ssscc_c*" from path
            ctd_plots._intermediate_residual_plot(
                btl_df.loc[btl_rows, cfg.column["refC"]]
                - btl_df.loc[btl_rows, cfg.column[cN]],
                btl_df.loc[btl_rows, cfg.column["p"]],
                btl_df.loc[btl_rows, "SSSCC"],
                xlabel=f"{cN.upper()} Residual (mS/cm)",
                f_out=f"{cfg.fig_dirs[cN]}residual_{f_stem}_prefit.pdf",
            )

            # 2) prepare data for fitting
            # NOTE: df_bad will be overwritten during post-fit data flagging
            # but is left here for future debugging (if necessary)
            df_good, df_bad = _prepare_fit_data(
                btl_df[good_rows],
                cfg.column[cN],
                cfg.column["refC"],
                zRange=fit_yaml[cN][f_stem]["zRange"],
            )
            ctd_plots._intermediate_residual_plot(
                df_good["Diff"],
                df_good[cfg.column["p"]],
                df_good["SSSCC"],
                xlabel=f"{cN.upper()} Residual (mS/cm)",
                f_out=f"{cfg.fig_dirs[cN]}residual_{f_stem}_fit_data.pdf",
            )

            # 3) calculate fit coefs
            # TODO: truncate coefs (10 digits? look at historical data)
            P_order = fit_yaml[cN][f_stem]["P_order"]
            T_order = fit_yaml[cN][f_stem]["T_order"]
            C_order = fit_yaml[cN][f_stem]["C_order"]
            coef_dict = multivariate_fit(
                df_good["Diff"],
                (df_good[cfg.column["p"]], P_order),
                (df_good[cfg.column[tN]], T_order),
                (df_good[cfg.column[cN]], C_order),
                coef_names=["cp", "ct", "cc"],
            )

            # 4) apply fit
            P_coefs = tuple(coef_dict[f"cp{n}"] for n in np.arange(1, P_order + 1))
            T_coefs = tuple(coef_dict[f"ct{n}"] for n in np.arange(1, T_order + 1))
            C_coefs = tuple(coef_dict[f"cc{n}"] for n in np.arange(1, C_order + 1))
            btl_df.loc[btl_rows, cfg.column[cN]] = apply_polyfit(
                btl_df.loc[btl_rows, cfg.column[cN]],
                (coef_dict["c0"],) + C_coefs,
                (btl_df.loc[btl_rows, cfg.column["p"]], P_coefs),
                (btl_df.loc[btl_rows, cfg.column[tN]], T_coefs),
            )
            time_df.loc[time_rows, cfg.column[cN]] = apply_polyfit(
                time_df.loc[time_rows, cfg.column[cN]],
                (coef_dict["c0"],) + C_coefs,
                (time_df.loc[time_rows, cfg.column["p"]], P_coefs),
                (time_df.loc[time_rows, cfg.column[tN]], T_coefs),
            )

            # 4.5) flag CTDCOND and make residual plots
            df_ques, df_bad = _flag_btl_data(
                btl_df[btl_rows],
                param=cfg.column[cN],
                ref=cfg.column["refC"],
                f_out=f"{cfg.fig_dirs[cN]}residual_{f_stem}.pdf",
            )

            # 5) handle quality flags
            C_flag = pd.concat([C_flag, df_bad, df_ques])

            # 6) handle fit params
            coef_df = pd.DataFrame()
            coef_df["SSSCC"] = ssscc_sublist
            coef_names = ["cp2", "cp1", "ct2", "ct1", "cc2", "cc1", "c0"]
            coef_df[coef_names] = 0.0
            for k, v in coef_dict.items():
                coef_df[k] = v

            C_fit_coefs = pd.concat([C_fit_coefs, coef_df])

        # one more fig with all cuts
        ctd_plots._intermediate_residual_plot(
            btl_df[cfg.column["refC"]] - btl_df[cfg.column[cN]],
            btl_df[cfg.column["p"]],
            btl_df["SSSCC"],
            xlabel=f"{cN.upper()} Residual (mS/cm)",
            show_thresh=True,
            f_out=f"{cfg.fig_dirs[cN]}residual_all_postfit.pdf",
        )

        # export cond quality flags
        # TODO: make these flags useful/less cluttered
        C_flag.sort_index().to_csv(f"{cfg.dirs['logs']}qual_flag_{cN}.csv", index=False)

        # export cond fit params
        C_fit_coefs[coef_names] = C_fit_coefs[coef_names].applymap(
            lambda x: np.format_float_scientific(x, precision=4, exp_digits=1)
        )
        C_fit_coefs.to_csv(cfg.dirs["logs"] + f"fit_coef_{cN}.csv", index=False)

    # recalculate salinity with calibrated C/T
    # TODO: compute CTDSAL1 and *2? how to decide which to use
    time_df[cfg.column["sal"]] = gsw.SP_from_C(
        time_df[cfg.column["c1"]],
        time_df[cfg.column["t1"]],
        time_df[cfg.column["p"]],
    )
    btl_df[cfg.column["sal"]] = gsw.SP_from_C(
        btl_df[cfg.column["c1"]],
        btl_df[cfg.column["t1"]],
        btl_df[cfg.column["p"]],
    )

    # flag salinity data
    # TODO: flag time using handcoded salts somehow? discrete vs continuous
    time_df[cfg.column["sal"] + "_FLAG_W"] = 2
    btl_df[cfg.column["sal"] + "_FLAG_W"] = flagging.by_residual(
        btl_df[cfg.column["sal"]],
        btl_df["SALNTY"],
        btl_df[cfg.column["p"]],
    )
    bad_rows = btl_df["SALNTY_FLAG_W"].isin([3, 4])
    btl_df.loc[bad_rows, cfg.column["sal"] + "_FLAG_W"] = 2  # bad salts not used for QC
    btl_df[cfg.column["sal"] + "_FLAG_W"] = flagging.nan_values(
        btl_df[cfg.column["sal"]], old_flags=btl_df[cfg.column["sal"] + "_FLAG_W"]
    )

    return btl_df, time_df



import os
import sys

import gsw
import numpy as np
import pandas as pd

"""code_pruning: is there anything worth keeping in this module?
only imported in outdated scripts"""


def string_converter(value):
    """To deal with Courtney CTD codes"""
    return value.split(":")[-1].strip()


def int_converter(value):
    """To deal with Courtney CTD codes"""
    return int(float(value.split(":")[-1].strip()))


def float_converter(value):
    """To deal with Courtney CTD codes"""
    return float(value.split(":")[-1])


def merge_courtney_ctd_codes(bottle_df, qc_df):
    """Take in ctd bottle file to be given to database, qc logs, and combine together.
    For use with Courtney style codes (P06W 2017).
    Input:
    bottle_file - dataframe
    qc_file - dataframe

    Output:
    b_test - dataframe with updated codes
    """

    flags = {
        "T": "REFTMP_FLAG_W",
        "T1": "CTDTMP1_FLAG_W",
        "T2": "CTDTMP2_FLAG_W",
        "C": "SALNTY_FLAG_W",
        "C1": "CTDCOND1_FLAG_W",
        "C2": "CTDCOND2_FLAG_W",
    }

    index = ["STNNBR", "CASTNO", "SAMPNO"]

    ### too lazy to fix it all in jupyter, do later
    a = qc_df
    b = bottle_df

    ### setup indicies for writing to
    a["CASTNO"] = a["stacast"] % 100
    a["STNNBR"] = (a["stacast"] - a["CASTNO"]) // 100
    a["new_params"] = None

    ### update new_params to flag_parameter for df.groupby() later
    for param, flag_param in flags.items():
        a.loc[a["parameter"] == param, "new_params"] = flag_param
    a_gb = a.groupby("new_params")

    ### df.update() requires the index to be set before being called
    b_test = b.set_index(index)

    ### loop through all flag_parameter names
    for param, flag_param in flags.items():
        try:
            a_gb_project = a_gb.get_group(flag_param).copy()
            a_gb_project.rename(columns={"flag": flag_param}, inplace=True)

            ### df.update() requires the index to be set before being called
            a_test = a_gb_project[index + [flag_param]].set_index(index)

            ### df.update() returns inplace, so return signature = None
            b_test.update(a_test)
        except KeyError:
            print(f"Parameter not found: {flag_param}")

    ### reset to work with again
    b_test = b_test.reset_index()

    return b_test


def load_courtney_ctd_codes(qc_file):
    """Load Courtney style ctd codes, and pass out dataframe.
    Input:
    qc_file - filepath to a courtney style ctd code
    Output:
    dataframe
    """
    converters = {
        0: int_converter,
        1: int_converter,
        2: string_converter,
        3: float_converter,
        4: int_converter,
        5: float_converter,
        6: float_converter,
        7: float_converter,
    }
    names = [
        "stacast",
        "SAMPNO",
        "parameter",
        "pressure",
        "flag",
        "primary_diff",
        "secondary_diff",
        "p_minus_s",
    ]
    return pd.read_csv(qc_file, header=None, converters=converters, names=names)


def load_to_odf_db_bottle_file(bottle_file):
    """File to be given to odf db maintainer to be loaded
    Input:
    bottle_file - filepath
    Output:
    dataframe
    """
    return pd.read_csv(bottle_file, skiprows=[1], skipfooter=1, engine="python")


def load_exchange_bottle_file(bottle_file):
    """WHP-exchange bottle file from odf db
    Input:
    bottle_file - filepath
    Output:
    dataframe
    """
    return pd.read_csv(
        bottle_file,
        skiprows=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11],
        skipfooter=1,
        engine="python",
    )


def write_to_odf_db_bottle_file(df_bottle, output_file):
    """Write WHP_exchange file for odf_db to load.
    Input:
    df_bottle - dataframe with bottlefile to write out
    output_file - filepath to write to
    Output:
    None
    """
    stringy = "CASTNO,CTDBACKSCATTER,CTDBACKSCATTER_FLAG_W,CTDFLUOR,CTDFLUOR_FLAG_W,"
    stringy += (
        "CTDOXY,CTDOXY_FLAG_W,CTDPRS,CTDPRS_FLAG_W,CTDRINKO,CTDRINKO_FLAG_W,CTDSAL,"
    )
    stringy += "CTDSAL_FLAG_W,CTDTMP,CTDTMP_FLAG_W,CTDXMISS,CTDXMISS_FLAG_W,REFTMP,REFTMP_FLAG_W,"
    stringy += "SAMPNO,STNNBR\n,VOLTS,,VOLTS,,UMOL/KG,,DBAR,,VOLTS,,PSS-78,,ITS-90,,VOLTS,,ITS-90,,,\n"
    stringy += df_bottle.to_string(
        header=False,
        index=False,
        formatters=[
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x) + ",",
            lambda x: str(x),
        ],
    )
    stringy += "\nEND_DATA\n"

    with open(output_file, "w") as f_handle:
        f_handle.write(stringy)
    return None


def merge_ctd_codes(log_folder, df_bottle):
    """
    Chopped version of old merge_ctd_codes to deal with dfs
    Input:
    log_folder - file path
    bottle_file - df

    Output:
    btl -  dataframe
    """
    ### Expecting all codes to be in one directory for now. Must change later
    log_files = os.listdir(log_folder)

    ### Merge quality codes, and store a copy in-memory in btl
    for f in log_files:
        x = load_courtney_ctd_codes(f"{log_folder}{f}")
        df_bottle = merge_courtney_ctd_codes(df_bottle, x)

    return df_bottle


def merge_bottle_trip_dfs(file_ssscc):
    """Merge the bottle trip dataframes, and add flag columns.
    Input:
    file_ssscc - file path
    Output:
    df_all - dataframe, holding all relevant dataframe info
    """
    df_all = pd.DataFrame()

    ### load in all station/casts to be used for ctd processing
    all_casts = []
    with open(file_ssscc, "r") as filename:
        all_casts = [line.strip() for line in filename]

    for ssscc in all_casts:
        station = int(ssscc[0:3])
        cast = int(ssscc[3:5])

        ### rewrite this line to be more portable
        df = pd.read_pickle(f"data/bottle/{ssscc}_btl_mean.pkl")

        ### chop dataframe shorter for ease of use
        df = df[
            ["CTDPRS", "CTDTMP1", "CTDTMP2", "CTDCOND1", "CTDCOND2", "btl_fire_num"]
        ]
        df.rename(index=str, columns={"btl_fire_num": "SAMPNO"}, inplace=True)

        ### add columns
        df["CASTNO"] = cast
        df["STNNBR"] = station
        df["CTDTMP1_FLAG_W"] = 2
        df["CTDTMP2_FLAG_W"] = 2
        df["CTDCOND1_FLAG_W"] = 2
        df["CTDCOND2_FLAG_W"] = 2

        ### finally, join the frames together to have one giant dataframe
        df_all = pd.concat([df_all, df], axis=0)

    return df_all


def prelim_ctd_bottle_df(file_ssscc, file_whp_bottle):
    """Merge bottle trip and bottle ctd info."""
    ### load bottle trip data
    df_bottle_trip = merge_bottle_trip_dfs(file_ssscc)
    ### load whp_bottle file from odf_db
    df_whp_bottle = load_exchange_bottle_file(file_whp_bottle)
    df_whp_bottle = df_whp_bottle[
        [
            "STNNBR",
            "CASTNO",
            "SAMPNO",
            "SALNTY",
            "SALNTY_FLAG_W",
            "CTDSAL",
            "CTDSAL_FLAG_W",
            "REFTMP",
            "REFTMP_FLAG_W",
            "CTDOXY",
            "CTDOXY_FLAG_W",
            "OXYGEN",
            "OXYGEN_FLAG_W",
        ]
    ]

    ### merge both dataframes together
    df_merged = df_bottle_trip.merge(df_whp_bottle, on=["STNNBR", "CASTNO", "SAMPNO"])
    return df_merged


def ctd_residuals_df(df_bottle):
    """Compute residuals and add to dataframe.
    Operate in place.
    """
    # Salinity should really be grabbed straight from the files, but...
    try:
        df_bottle["BTLCOND"] = gsw.C_from_SP(
            df_bottle["SALNTY"], df_bottle["CTDTMP1"], df_bottle["CTDPRS"]
        )
    except ValueError:
        print("WHOOPS SOMETHING WENT WRONG")
        df_bottle["SALNTY"] = df_bottle["SALNTY"].replace(to_replace=-999, value=np.nan)
        df_bottle["BTLCOND"] = gsw.C_from_SP(
            df_bottle["SALNTY"], df_bottle["CTDTMP1"], df_bottle["CTDPRS"]
        )

    df_bottle["BTL_O"] = df_bottle["OXYGEN"] - df_bottle["CTDOXY"]

    df_bottle["BTL_C1"] = df_bottle["BTLCOND"] - df_bottle["CTDCOND1"]
    df_bottle["BTL_C2"] = df_bottle["BTLCOND"] - df_bottle["CTDCOND2"]
    df_bottle["C1_C2"] = df_bottle["CTDCOND1"] - df_bottle["CTDCOND2"]

    df_bottle["BTL_T1"] = df_bottle["REFTMP"] - df_bottle["CTDTMP1"]
    df_bottle["BTL_T2"] = df_bottle["REFTMP"] - df_bottle["CTDTMP2"]
    df_bottle["T1_T2"] = df_bottle["CTDTMP1"] - df_bottle["CTDTMP2"]

    df_bottle["BTL_SAL_UP"] = df_bottle["SALNTY"] - gsw.SP_from_C(
        df_bottle["CTDCOND2"], df_bottle["CTDTMP2"], df_bottle["CTDPRS"]
    )
    df_bottle["BTL_SAL"] = df_bottle["SALNTY"] - df_bottle["CTDSAL"]
    return None


def get_qc_t_from_df(df):
    """Limit data to only good data (WOCE discrete sample code 2)
    Returns a new copy of a dataframe.
    Input:
    df - dataframe of all data
    Output:
    df limited to only good temperature data
    """
    df = (
        df.groupby(["REFTMP_FLAG_W", "CTDTMP1_FLAG_W", "CTDTMP2_FLAG_W"])
        .get_group((2, 2, 2))
        .copy()
    )
    return df


def get_qc_c_from_df(df):
    """Limit data to only good data (WOCE discrete sample code 2)
    Returns a new copy of a dataframe.
    Input:
    df - dataframe of all data
    Output:
    df limited to only good conductivity data
    """
    df = (
        df.groupby(["SALNTY_FLAG_W", "CTDCOND1_FLAG_W", "CTDCOND2_FLAG_W"])
        .get_group((2, 2, 2))
        .copy()
    )
    return df


def get_qc_s_from_df(df):
    """Limit data to only good data (WOCE discrete sample code 2)
    Returns a new copy of a dataframe.
    Input:
    df - dataframe of all data
    Output:
    df limited to only good salinity data
    """
    df = df.groupby(["SALNTY_FLAG_W", "CTDSAL_FLAG_W"]).get_group((2, 2)).copy()
    return df


def get_qc_o_from_df(df):
    """Limit data to only good data (WOCE discrete sample code 2)
    Returns a new copy of a dataframe.
    Input:
    df - dataframe of all data
    Output:
    df limited to only good oxygen data
    """
    df = df.groupby(["OXYGEN_FLAG_W", "CTDOXY_FLAG_W"]).get_group((2, 2)).copy()
    return df


def get_qc_all_from_df(df):
    """Limit data to only good data (WOCE discrete sample code 2)
    Returns a new copy of a dataframe.
    Input:
    df - dataframe of all data
    Output:
    df limited to only good ctdo data
    """
    try:
        df = (
            df.groupby(
                [
                    "REFTMP_FLAG_W",
                    "CTDTMP1_FLAG_W",
                    "CTDTMP2_FLAG_W",
                    "CTDCOND1_FLAG_W",
                    "CTDCOND2_FLAG_W",
                    "SALNTY_FLAG_W",
                    "CTDSAL_FLAG_W",
                    "OXYGEN_FLAG_W",
                    "CTDOXY_FLAG_W",
                ]
            )
            .get_group((2, 2, 2, 2, 2, 2, 2, 2, 2))
            .copy()
        )
    except KeyError:
        df = (
            df.groupby(
                [
                    "REFTMP_FLAG_W",
                    "CTDTMP1_FLAG_W",
                    "CTDTMP2_FLAG_W",
                    "CTDCOND1_FLAG_W",
                    "CTDCOND2_FLAG_W",
                    "SALNTY_FLAG_W",
                    "CTDSAL_FLAG_W",
                    "OXYGEN_FLAG_W",
                    "CTDOXY_FLAG_W",
                ]
            )
            .get_group((2, 2, 2, 2, 2, 2, 2, 2, 1))
            .copy()
        )
    return df


def residual_stddev(
    df,
    param=[
        "BTL_O",
        "BTL_C1",
        "BTL_C2",
        "C1_C2",
        "BTL_T1",
        "BTL_T2",
        "T1_T2",
        "BTL_SAL",
        "BTL_SAL_UP",
    ],
):
    """Calculate standard deviations of parameters or residuals.
    Input:
    df - dataframe containing all data from casts
    param - list of parameter names
    Output:
    output - dictionary, key = is parameter name or parameter deep, value is 2 stddevs
    """
    output = {}
    for x in param:
        output[f"{x}"] = df[f"{x}"].std() * 2
        df_deep = df[df["CTDPRS"] > 2000]
        output[f"{x}_DEEP"] = df_deep[f"{x}"].std() * 2
    return output


def main(argv):
    """Example run"""
    qual_codes_filepath = "data/quality_codes/"
    cruise_dir = "data/"
    log_dir = "data/logs"

    file_ssscc = f"{cruise_dir}ssscc.csv"
    bottle_file = f"{qual_codes_filepath}320620180309_hy1.csv"

    ### compile bottle trips into one file, then merge with odf_db bottle file
    df_bottle = prelim_ctd_bottle_df(file_ssscc, bottle_file)
    ### compute residuals for plotting
    ctd_residuals_df(df_bottle)
    load_courtney_ctd_codes()
    df_bottle = merge_ctd_codes(log_dir, df_bottle)
    df_bottle.to_pickle(f"{qual_codes_filepath}merged_ctd_codes.pkl")

    # df_coded = get_qc_all_from_df(df_bottle)
    ### write file out
    return None


if __name__ == "__main__":
    main(sys.argv[1:])



import logging
import numpy as np
import pandas as pd

from .. import ctd_plots, get_ctdcal_config

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)

btl_file = "data/scratch_folder/report_data.csv"
btl_df = pd.read_csv(btl_file)


def plot_residuals(outdir="data/report_figs/", ext=".pdf"):

    # temperature plots
    log.info("Generating temperature residual plots")
    ctd_plots.residual_vs_pressure(
        btl_df["CTDTMP1"],
        btl_df["REFTMP"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlabel="CTDTMP1 Residual (T90 C)",
        f_out=f"{outdir}reft-t1_vs_p{ext}",
    )
    ctd_plots.residual_vs_pressure(
        btl_df["CTDTMP2"],
        btl_df["REFTMP"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlabel="CTDTMP2 Residual (T90 C)",
        f_out=f"{outdir}reft-t2_vs_p{ext}",
    )
    ctd_plots.residual_vs_pressure(
        btl_df["CTDTMP2"],
        btl_df["CTDTMP1"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlabel="CTDTMP1-CTDTMP2 Residual (T90 C)",
        f_out=f"{outdir}t1-t2_vs_p{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDTMP1"],
        btl_df["REFTMP"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDTMP1 Residual (T90 C)",
        f_out=f"{outdir}reft-t1_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDTMP2"],
        btl_df["REFTMP"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDTMP2 Residual (T90 C)",
        f_out=f"{outdir}reft-t2_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDTMP2"],
        btl_df["CTDTMP1"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDTMP1-CTDTMP2 Residual (T90 C)",
        f_out=f"{outdir}t1-t2_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDTMP1"],
        btl_df["REFTMP"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDTMP1 Residual (T90 C)",
        deep=True,
        f_out=f"{outdir}reft-t1_vs_stn_deep{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDTMP2"],
        btl_df["REFTMP"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDTMP2 Residual (T90 C)",
        deep=True,
        f_out=f"{outdir}reft-t2_vs_stn_deep{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDTMP1"],
        btl_df["CTDTMP2"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDTMP1-CTDTMP2 Residual (T90 C)",
        deep=True,
        f_out=f"{outdir}t1-t2_vs_stn_deep{ext}",
    )

    # conductivity plots
    log.info("Generating conductivity residual plots")
    ctd_plots.residual_vs_pressure(
        btl_df["CTDCOND1"],
        btl_df["BTLCOND"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlabel="CTDCOND1 Residual (mS/cm)",
        f_out=f"{outdir}refc-c1_vs_p{ext}",
    )
    ctd_plots.residual_vs_pressure(
        btl_df["CTDCOND2"],
        btl_df["BTLCOND"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlabel="CTDCOND2 Residual (mS/cm)",
        f_out=f"{outdir}refc-c2_vs_p{ext}",
    )
    ctd_plots.residual_vs_pressure(
        btl_df["CTDCOND1"],
        btl_df["CTDCOND2"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlabel="CTDCOND1-CTDCOND2 Residual (mS/cm)",
        f_out=f"{outdir}c1-c2_vs_p{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDCOND1"],
        btl_df["BTLCOND"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDCOND1 Residual (mS/cm)",
        f_out=f"{outdir}refc-c1_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDCOND2"],
        btl_df["BTLCOND"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDCOND2 Residual (mS/cm)",
        f_out=f"{outdir}refc-c2_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDCOND1"],
        btl_df["CTDCOND2"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDCOND1-CTDCOND2 Residual (mS/cm)",
        f_out=f"{outdir}c1-c2_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDCOND1"],
        btl_df["BTLCOND"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDCOND1 Residual (mS/cm)",
        deep=True,
        f_out=f"{outdir}refc-c1_vs_stn_deep{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDCOND2"],
        btl_df["BTLCOND"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDCOND2 Residual (mS/cm)",
        deep=True,
        f_out=f"{outdir}refc-c2_vs_stn_deep{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDCOND1"],
        btl_df["CTDCOND2"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDCOND1-CTDCOND2 Residual (mS/cm)",
        deep=True,
        f_out=f"{outdir}c1-c2_vs_stn_deep{ext}",
    )
    # coherence plot doesn't have its own function...
    import matplotlib.pyplot as plt

    plt.figure(figsize=(7, 6))
    plt.scatter(
        btl_df["CTDTMP1"] - btl_df["CTDTMP2"],
        btl_df["CTDCOND1"] - btl_df["CTDCOND2"],
        c=btl_df["CTDPRS"],
        marker="+",
    )
    plt.xticks(rotation=45)
    cbar = plt.colorbar(pad=0.1)
    cbar.set_label("Pressure (dbar)")
    plt.xlim((-0.05, 0.05))
    plt.ylim((-0.05, 0.05))
    plt.xlabel("CTDTMP1-CTDTMP2 Residual (T90 C)", fontsize=12)
    plt.ylabel("CTDCOND1-CTDCOND2 Residual (mS/cm)", fontsize=12)
    plt.title("CTDCOND1-CTDCOND2 vs. CTDTMP1-CTDTMP2", fontsize=12)
    plt.tight_layout()
    plt.savefig(f"{outdir}c_t_coherence{ext}")
    plt.close()

    # salinity plots
    log.info("Generating salinity residual plots")
    ctd_plots.residual_vs_pressure(
        btl_df["CTDSAL"],
        btl_df["SALNTY"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlabel="CTDSAL Residual (PSU)",
        f_out=f"{outdir}btlsal-sal_vs_p{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDSAL"],
        btl_df["SALNTY"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDSAL Residual (PSU)",
        f_out=f"{outdir}btlsal-sal_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDSAL"],
        btl_df["SALNTY"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylabel="CTDSAL Residual (PSU)",
        deep=True,
        f_out=f"{outdir}btlsal-sal_vs_stn_deep{ext}",
    )

    # SBE43 oxygen plots
    log.info("Generating oxygen (SBE43) residual plots")
    ctd_plots.residual_vs_pressure(
        btl_df["CTDOXY"],
        btl_df["OXYGEN"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlim=(-10, 10),
        xlabel="CTDOXY Residual (umol/kg)",
        f_out=f"{outdir}oxy-43_vs_p{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDOXY"],
        btl_df["OXYGEN"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylim=(-10, 10),
        ylabel="CTDOXY Residual (umol/kg)",
        f_out=f"{outdir}oxy-43_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDOXY"],
        btl_df["OXYGEN"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        deep=True,
        ylim=(-10, 10),
        ylabel="CTDOXY Residual (umol/kg)",
        f_out=f"{outdir}oxy-43_vs_stn_deep{ext}",
    )
    # ctd_plots.btl_oxy_residuals_temperature_plot(btl_df)
    # ctd_plots.btl_oxy_residuals_station_temperature_plot(btl_df)
    # ctd_plots.btl_oxy_residuals_station_deep_temperature_plot(btl_df)
    # ctd_plots.btl_oxy_residuals_pressure_concentration_plot(btl_df)
    # ctd_plots.btl_oxy_residuals_station_concentration_plot(btl_df)

    # RINKO oxygen plots
    log.info("Generating oxygen (RINKO) residual plots")
    ctd_plots.residual_vs_pressure(
        btl_df["CTDRINKO"],
        btl_df["OXYGEN"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        xlim=(-10, 10),
        xlabel="CTDRINKO Residual (umol/kg)",
        f_out=f"{outdir}oxy-rinko_vs_p{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDRINKO"],
        btl_df["OXYGEN"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        ylim=(-10, 10),
        ylabel="CTDRINKO Residual (umol/kg)",
        f_out=f"{outdir}oxy-rinko_vs_stn{ext}",
    )
    ctd_plots.residual_vs_station(
        btl_df["CTDRINKO"],
        btl_df["OXYGEN"],
        btl_df["CTDPRS"],
        btl_df["STNNBR"],
        deep=True,
        ylim=(-10, 10),
        ylabel="CTDRINKO Residual (umol/kg)",
        f_out=f"{outdir}oxy-rinko_vs_stn_deep{ext}",
    )


def pressure_offset():

    data = pd.read_csv("data/logs/ondeck_pressure.csv")
    print(f"Average deck pressure:\n{data.describe().loc[['min', 'max', 'mean']]}\n")
    print(f"Average offset:\n{(data['ondeck_end_p'] - data['ondeck_start_p']).describe()}\n")


def fit_coefficients():
    """Write code to build table for fit groups?"""
    pass


def calculate_residuals():

    log.info("Calculating T/C/O residuals\n")

    low_grad_rows = (btl_df["CTDTMP1"] - btl_df["CTDTMP2"]).abs() < 0.002
    deep_rows = btl_df["CTDPRS"] > 2000

    T_flag2 = (btl_df["CTDTMP1_FLAG_W"] == 2) | (btl_df["CTDTMP2_FLAG_W"] == 2)
    C_flag2 = (btl_df["CTDCOND1_FLAG_W"] == 2) | (btl_df["CTDCOND2_FLAG_W"] == 2)
    S_flag2 = (btl_df["CTDSAL_FLAG_W"] == 2) | (btl_df["CTDSAL_FLAG_W"] == 2)
    O_flag2 = btl_df["CTDOXY_FLAG_W"] == 2
    R_flag2 = btl_df["CTDRINKO_FLAG_W"] == 2

    def fmt(x):
        return np.format_float_positional(x, precision=5)

    # Temperature
    low_grad = btl_df[low_grad_rows & T_flag2]
    deep = btl_df[deep_rows & T_flag2]
    print(f"REFT-T1 = {fmt((low_grad['REFTMP'] - low_grad['CTDTMP1']).std() * 2)}")
    print(f"REFT-T2 = {fmt((low_grad['REFTMP'] - low_grad['CTDTMP2']).std() * 2)}")
    print(f"T1-T2 = {fmt((low_grad['CTDTMP1'] - low_grad['CTDTMP2']).std() * 2)}")
    print(f"REFT-T1 (deep) = {fmt((deep['REFTMP'] - deep['CTDTMP1']).std() * 2)}")
    print(f"REFT-T2 (deep) = {fmt((deep['REFTMP'] - deep['CTDTMP2']).std() * 2)}")
    print(f"T1-T2 (deep) = {fmt((deep['CTDTMP1'] - deep['CTDTMP2']).std() * 2)}")
    print("")

    # Conductivity
    low_grad = btl_df[low_grad_rows & C_flag2]
    deep = btl_df[deep_rows & C_flag2]
    print(f"REFC-C1 = {fmt((low_grad['BTLCOND'] - low_grad['CTDCOND1']).std() * 2)}")
    print(f"REFC-C2 = {fmt((low_grad['BTLCOND'] - low_grad['CTDCOND2']).std() * 2)}")
    print(f"C1-C2 = {fmt((low_grad['CTDCOND1'] - low_grad['CTDCOND2']).std() * 2)}")
    print(f"REFC-C1 (deep) = {fmt((deep['BTLCOND'] - deep['CTDCOND1']).std() * 2)}")
    print(f"REFC-C2 (deep) = {fmt((deep['BTLCOND'] - deep['CTDCOND2']).std() * 2)}")
    print(f"C1-C2 (deep) = {fmt((deep['CTDCOND1'] - deep['CTDCOND2']).std() * 2)}")
    print("")

    # Salinity
    low_grad = btl_df[low_grad_rows & S_flag2]
    deep = btl_df[deep_rows & S_flag2]
    print(f"REFS-SAL = {fmt((low_grad['SALNTY'] - low_grad['CTDSAL']).std() * 2)}")
    print(f"REFS-SAL (deep) = {fmt((deep['SALNTY'] - deep['CTDSAL']).std() * 2)}")
    print("")

    # SBE43
    # low_grad = btl_df[low_grad_rows & O_flag2]
    flag2 = btl_df[O_flag2]
    deep = btl_df[deep_rows & O_flag2]
    print(f"OXY-SBE43 = {fmt((flag2['OXYGEN'] - flag2['CTDOXY']).std() * 2)}")
    print(f"OXY-SBE43 (deep) = {fmt((deep['OXYGEN'] - deep['CTDOXY']).std() * 2)}")
    print("")

    # RINKO
    # low_grad = btl_df[low_grad_rows & R_flag2]
    flag2 = btl_df[R_flag2]
    deep = btl_df[deep_rows & R_flag2]
    print(f"OXY-RINKO = {fmt((flag2['OXYGEN'] - flag2['CTDRINKO']).std() * 2)}")
    print(f"OXY-RINKO (deep) = {fmt((deep['OXYGEN'] - deep['CTDRINKO']).std() * 2)}")
    print("")


def cruise_report_residuals():
    """Do all the things for cruise report"""
    plot_residuals()
    pressure_offset()
    fit_coefficients()
    calculate_residuals()


# if __name__ == "__main__":
#     make_cruise_report_plots()



#!/usr/bin/env python
import csv
import logging
import math
import warnings
from collections import OrderedDict
from datetime import datetime, timezone
from pathlib import Path

import gsw
import numpy as np
import pandas as pd
import scipy.signal as sig

from . import flagging as flagging
from . import get_ctdcal_config
from . import oxy_fitting as oxy_fitting
from . import report_ctd as report_ctd

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)

warnings.filterwarnings("ignore", "Mean of empty slice.")


def cast_details(df, ssscc, log_file=None):
    """
    We determine the cast details using pandas magic.
    First find alternating periods of pumps on and pumps off, then select the
    pumps on period with the highest pressure. Get values from the row with the
    highest pressure, and return all values to be sent to log.

    Parameters
    ----------
    df : DataFrame
        Filtered CTD data
    ssscc : integer
        The station and cast, as SSSCC format
    log_file : file handle or string
        File destination for cast details

    Returns
    -------
    df_downcast : DataFrame
        CTD data with the soak period and upcast trimmed off

    Notes
    -----
    The following (float) variables are output to log_file:
    time_start : Time at start of cast (in unix epoch time)
    time_end : Time at end of cast (in unix epoch time)
    time_bottom : Time at bottom of cast (in unix epoch time)
    p_start : Pressure at which cast started
    p_max : Bottom of the cast pressure
    b_lat : Latitude at bottom of cast
    b_lon : Longitude at bottom of cast
    b_alt : Altimeter reading at bottom of cast
    """
    df_cast = _trim_soak_period(df)

    # TODO: call parameters from config file instead
    p_start = float(np.around(df_cast["CTDPRS"].head(1), 4))
    p_max_ind = df_cast["CTDPRS"].argmax()
    p_max = float(np.around(df_cast["CTDPRS"].max(), 4))
    time_start = float(df_cast["scan_datetime"].head(1))
    time_end = float(df_cast["scan_datetime"].tail(1))
    time_bottom = float(df_cast["scan_datetime"][p_max_ind])
    b_lat = float(np.around(df_cast["GPSLAT"][p_max_ind], 4))
    b_lon = float(np.around(df_cast["GPSLON"][p_max_ind], 4))
    b_alt = float(np.around(df_cast["ALT"][p_max_ind], 4))

    report_ctd.report_cast_details(
        ssscc,
        log_file,
        time_start,
        time_end,
        time_bottom,
        p_start,
        p_max,
        b_alt,
        b_lat,
        b_lon,
    )

    # remove upcast
    df_downcast = df_cast[:p_max_ind].copy()

    return df_downcast


def _trim_soak_period(df=None):
    """
    1) Find pump on/off patterns
    2) Select pump_on=True group with largest pressure recording
    3) Find soak period before start of downcast
    4) Trim cast, return everything after top of cast (i.e. minimum pressure)
    """
    df_list = [
        g for i, g in df.groupby(df["pump_on"].ne(df["pump_on"].shift()).cumsum())
    ]
    df_pump_on_list = [df for df in df_list if df["pump_on"].all()]
    df_cast = df_pump_on_list[np.argmax([df["CTDPRS"].max() for df in df_pump_on_list])]
    df_cast = df_cast.reset_index(drop=True)
    # next fn deals w/ edge cases, leave as is for now
    df_cast = _find_last_soak_period(df_cast)
    start_ind = df_cast.loc[: len(df) // 4, "CTDPRS"].argmin()
    df_trimmed = df_cast[start_ind:].reset_index(drop=True).copy()

    return df_trimmed


def _find_last_soak_period(df_cast, time_bin=8, P_surface=2, P_downcast=50):
    """
    Find the soak period before the downcast starts.

    The algorithm is tuned for repeat hydrography work, specifically US GO-SHIP
    parameters. This assumes the soak depth will be somewhere between 10 and 30
    meters, the package will sit at the soak depth for at least 20 to 30 seconds
    before starting ascent to the surface and descent to target depth.

    The algorithm is not guaranteed to catch the exact start of the soak period,
    but within a minimum period of time_bin seconds(?) from end of the soak if
    the soak period assumption is valid. This should be shorter than the total
    soak period time, and able to catch the following rise and descent of the
    package that signals the start of the cast.

    The algorithm has been designed to handle four general cases of casts:
        * A routine cast with pumps turning on in water and normal soak
        * A cast where the pumps turn on in air/on deck
        * A cast where the pumps turn on and off due to rosette coming out of water
        * A cast where there are multiple stops on the downcast to the target depth

    Parameters
    ----------
    df_cast : DataFrame
        DataFrame of the entire cast, from deckbox on to deckbox off
    time_bin : integer, optional
        Number of seconds to bin average for descent rate calculation
    P_surface : integer, optional
        Minimum surface pressure threshold required to look for soak depth
        (2 dbar was chosen as an average rosette is roughly 1.5 to 2 meters tall)
    P_downcast : integer, optional
        Minimum pressure threshold required to assume downcast has started
        (50 dbar has been chosen as double the deep soak depth of 20-30 dbar)

    Returns
    -------
    df_cast_trimmed : DataFrame
        DataFrame starting within time_bin seconds of the last soak period.
    """
    # Validate user input
    if time_bin <= 0:
        raise ValueError("Time bin value should be positive whole seconds.")
    if P_downcast <= 0:
        raise ValueError(
            "Starting downcast pressure threshold must be positive integers."
        )
    if P_downcast < P_surface:
        raise ValueError(
            "Starting downcast pressure threshold must be greater \
                        than surface pressure threshold."
        )

    # If pumps have not turned on until in water, return DataFrame
    if df_cast.iloc[0]["CTDPRS"] > P_surface:
        return df_cast

    # Bin the data by time, and compute the average rate of descent
    df_cast["index"] = df_cast.index  # needed at end to identify start_idx
    df_cast["bin"] = pd.cut(
        df_cast.index,
        np.arange(df_cast.index[0], df_cast.index[-1], time_bin * 24),
        labels=False,
        include_lowest=True,
    )
    df_binned = df_cast.groupby("bin").mean()

    # Compute difference of descent rates and label bins
    df_binned["dP"] = df_binned["CTDPRS"].diff().fillna(0).round(0)
    df_binned["movement"] = pd.cut(
        df_binned["dP"], [-1000, -0.5, 0.5, 1000], labels=["up", "stop", "down"]
    )

    # Find all periods where the rosette is not moving
    df_group = df_binned.groupby(
        df_binned["movement"].ne(df_binned["movement"].shift()).cumsum()
    )
    df_list = [g for i, g in df_group]

    # Find last soak period before starting descent to target depth
    def find_last(df_list, P_downcast):
        for idx, df in enumerate(df_list):
            if df["CTDPRS"].max() < P_downcast:
                # make sure it's soak, not a stop to switch to autocast (i.e. A20 2021)
                # TODO: try instead finding max depth then working backwards?
                if df.max()["movement"] == "stop" and len(df) > 1:
                    last_idx = idx
            else:
                return last_idx
        return last_idx

    # Trim off everything before last soak
    start_idx = int(df_list[find_last(df_list, P_downcast)].head(1)["index"])
    df_cast_trimmed = df_cast.loc[start_idx:].reset_index()

    return df_cast_trimmed


def ctd_align(inMat=None, col=None, time=0.0):
    """ctd_align function

    Function takes full NUMPY ndarray with predefined dtype array
    and adjusts time of sensor responce and water flow relative to
    the time frame of temperature sensor.

    Originally written by Courtney Schatzman, docstring by Joseph Gum.
    Need to generate alignment plots in order to properly use ctd_align.

    Args:
        param1 (ndarray): inMat, numpy ndarray with dtype array
        param2 (float): col, column to apply time advance to.
        param3 (float): time, advance in seconds to apply to raw data.

    Returns:
        Narray: The return value is ndarray with adjusted time of parameter
          specified.

    """
    # Num of frames per second.
    fl = 24

    if (inMat is not None) & (col is not None) & (time > 0.0):
        # Time to advance
        advnc = int(fl * time)
        tmp = np.arange(advnc, dtype=np.float)
        last = inMat[col][len(inMat) - 1]
        tmp.fill(float(last))
        inMat[col] = np.concatenate((inMat[col][advnc:], tmp))

    return inMat


def raw_ctd_filter(df=None, window="triangle", win_size=24, parameters=None):
    """
    Filter raw CTD data using one of three window types (boxcar, hanning, triangle).

    Parameters
    ----------
    df : DataFrame
        Raw CTD data
    window : str, optional
        Type of filter window
    win_size : int, optional
        Length of window in number of samples
    parameters : list of str, optional
        List of DataFrame columns to be filtered

    Returns
    -------
    filtered_df : DataFrame
        CTD data with filtered parameters
    """

    filter_df = df.copy()
    if parameters is not None:
        for p in parameters:
            if window == "boxcar":
                win = sig.boxcar(win_size)
            elif window == "hanning":
                win = sig.hann(win_size)
            elif window == "triangle":
                win = sig.triang(win_size)
            filter_df[p] = sig.convolve(filter_df[p], win, mode="same") / np.sum(win)

    return filter_df


def remove_on_deck(df, stacast, cond_startup=20.0, log_file=None):
    """
    Find and remove times when rosette is on deck.
    Optionally log average pressure at start and end of cast.

    Parameters
    ----------
    df : DataFrame
        Raw CTD data
    stacast : str
        Station/cast name
    cond_startup : float, optional
        Minimum conductivity (units?) threshold indicating rosette is in water
    log_file : str, optional
        Path and filename to save start/end deck pressure values

    Returns
    -------
    trimmed_df : DataFrame
        Raw CTD data trimmed to times when rosette is in water
    """
    # TODO: move these to config file
    # Frequency
    fl = 24
    fl2 = fl * 2
    # Half minute
    ms = 30
    time_delay = fl * ms  # time to let CTD pressure reading settle/sit on deck

    # split dataframe into upcast/downcast
    downcast = df.iloc[: (df["CTDPRS"].argmax() + 1)]
    upcast = df.iloc[(df["CTDPRS"].argmax() + 1) :]

    # Search each half of df for minimum conductivity
    # threshold to identify when rosette is out of water
    start_df = downcast.loc[
        (downcast[cfg.column["c1"]] < cond_startup)
        & (downcast[cfg.column["c2"]] < cond_startup),
        cfg.column["p"],
    ]
    end_df = upcast.loc[
        (upcast[cfg.column["c1"]] < cond_startup)
        & (upcast[cfg.column["c2"]] < cond_startup),
        cfg.column["p"],
    ]

    # Evaluate starting and ending pressures
    start_samples = len(start_df)
    if start_samples > time_delay:
        start_p = np.average(start_df.iloc[fl2 : (start_samples - time_delay)])
    else:
        start_seconds = start_samples / fl
        log.warning(
            f"{stacast}: Only {start_seconds:0.1f} seconds of start pressure averaged."
        )
        start_p = np.average(start_df.iloc[fl2:start_samples])

    end_samples = len(end_df)
    if end_samples > time_delay:
        end_p = np.average(end_df.iloc[(time_delay):])
    else:
        end_seconds = end_samples / fl
        log.warning(
            f"{stacast}: Only {end_seconds:0.1f} seconds of end pressure averaged."
        )
        end_p = np.average(end_df)  # just average whatever there is

    # Remove ondeck start and end pressures
    if len(start_df) == 0:
        log.warning("Failed to find starting deck pressure.")
        for n in [1, 2]:
            try:
                (downcast[cfg.column[f"c{n}"]] < cond_startup).value_counts()[True]
            except KeyError:
                log.warning(
                    f"No values below {cond_startup} found for {cfg.column[f'c{n}']}"
                )
        breakpoint()
    if len(end_df) == 0:
        log.warning("Failed to find ending deck pressure.")
        for n in [1, 2]:
            try:
                (upcast[cfg.column[f"c{n}"]] < cond_startup).value_counts()[True]
            except KeyError:
                log.warning(
                    f"No values below {cond_startup} found for {cfg.column[f'c{n}']}"
                )
        breakpoint()
    # MK (3/23/20, 11am):
    # auto end calculation failed bc cond2 is still >30
    # may have to do manually or just use cond1 for station 00901
    trimmed_df = df.iloc[start_df.index.max() : end_df.index.min()].copy()

    # Log ondeck pressures
    if log_file is not None:
        report_ctd.report_pressure_details(stacast, log_file, start_p, end_p)

    return trimmed_df


def roll_filter(df, p_col="CTDPRS", direction="down"):
    """
    Filter out heaving in CTD data due to ship rolls.

    Parameters
    ----------
    df : DataFrame
        CTD data
    p_col : str, optional
        Name of pressure column
    direction : str, optional
        Direction of cast (i.e. "down" or "up" cast)

    Returns
    -------
    var2 : dtype
        var description

    """
    if direction == "down":
        monotonic_sequence = df[p_col].expanding().max()
    elif direction == "up":
        monotonic_sequence = df[p_col].expanding().min()
    else:
        raise ValueError("direction must be one of (up, down)")

    return df[df[p_col] == monotonic_sequence]


def pressure_sequence(df, p_col="CTDPRS", direction="down"):
    """
    Convert CTD time series to a pressure series.

    Parameters
    ----------
    df : DataFrame
        CTD time series data
    p_col : str, optional
        Name of pressure column
    direction : str, optional
        Direction to sequence data

    Returns
    -------
    df_binned : DataFrame
        Pressure binned CTD data
    """
    # TODO: optional start/end pressure values?

    # change to take dataframe with the following properties
    # * in water data only (no need to find cast start/end)
    # * The full down and up time series (not already split since this method will do it)
    # New "algorithm" (TODO spell this right)
    # * if direction is "down", use the input as is
    # * if direction is "up", invert the row order of the input dataframe
    # Use the "roll filter" method to get only the rows to be binned
    # * the roll filter will treat the "up" part of the cast as a giant roll to be filtered out
    # * the reversed dataframe will ensure we get the "up" or "down" part of the cast
    # * there is no need to reverse the dataframe again as the pressure binning process will remove any "order" information (it doesn't care about the order)
    # That's basically all I (barna) have so far TODO Binning, etc...
    # pandas.cut() to do binning

    df_filtered = roll_filter(df, p_col, direction=direction)

    # 04/11/21 MK: this is not behaving properly or the order is wrong?
    # Currently this function fills the top-most good CTD value *before* bin avg,
    # so those bin averages are not the same as the first good binned value.
    # df_filled = _fill_surface_data(df_filtered, bin_size=2)

    df_binned = binning_df(df_filtered, bin_size=2)  # TODO: abstract bin_size in config
    fill_rows = df_binned["CTDPRS"].isna()
    df_binned.loc[fill_rows, "CTDPRS"] = df_binned[fill_rows].index.to_numpy()
    df_binned.bfill(inplace=True)
    df_binned.loc[:, "interp_bool"] = False
    df_binned.loc[fill_rows, "interp_bool"] = True
    df_filled = _flag_backfill_data(df_binned).drop(columns="interp_bool")

    return df_filled.reset_index(drop=True)


def binning_df(df, p_column="CTDPRS", bin_size=2):
    """Calculate the bin-mean of each column in input dataframe

    Parameters
    ----------
    df : DataFrame
        Data to be bin-meaned
    p_column : str, optional
        Pressure column name to use for binning
    bin_size : int, optional
        Width of bins (in decibars)

    Returns
    -------
    df_out : DataFrame
        Bin-meaned data

    """
    if p_column not in df.columns:
        raise KeyError(f"{p_column} column missing from dataframe")

    p_max = np.ceil(df[p_column].max())
    labels = np.arange(0, p_max, bin_size)
    bin_edges = np.arange(0, p_max + bin_size, bin_size)
    df_out = df.copy()
    df_out.loc[:, "bins"] = pd.cut(
        df[p_column], bins=bin_edges, right=False, include_lowest=True, labels=labels
    )
    df_out.loc[:, p_column] = df_out["bins"].astype(float)

    return df_out.groupby("bins").mean()


def _fill_surface_data(df, bin_size=2):
    """Copy first scan from top of cast and propagate up to surface at bin centers"""
    df = df.copy(deep=True)
    p_min = np.floor(df["CTDPRS"].iloc[0])
    df_surface = pd.DataFrame({"CTDPRS": np.arange(bin_size / 2, p_min, bin_size)})
    df_surface["interp_bool"] = True
    df = df_surface.merge(df, on="CTDPRS", how="outer")
    df["interp_bool"].fillna(False, inplace=True)
    df = _flag_backfill_data(df).drop(columns="interp_bool")

    return df.bfill()


def _get_pressure_offset(start_vals, end_vals):
    """
    Finds unique values and calculate mean for pressure offset

    Parameters
    ----------
    start_vals : array_like
        Array of initial ondeck pressure values

    end_vals : array_like
        Array of ending ondeck pressure values
    Returns
    -------
    p_off : float
         Average pressure offset

    """
    p_start = pd.Series(np.unique(start_vals))
    p_end = pd.Series(np.unique(end_vals))
    p_start = p_start[p_start.notnull()]
    p_end = p_end[p_end.notnull()]
    p_off = p_start.mean() - p_end.mean()

    # JACKSON THINKS THIS METHOD SHOULD BE USED TO KEEP START END PAIRS
    #    p_df = pd.DataFrame()
    #    p_df['p_start'] = p_start
    #    p_df['p_end'] = p_end
    #    p_df = p_df[p_df['p_end'].notnull()]
    #    p_df = p_df[p_df['p_start'].notnull()]
    #    p_off = p_df['p_start'].mean() - p_df['p_end'].mean()
    ##########################################################

    p_off = np.around(p_off, decimals=4)

    return p_off


def apply_pressure_offset(df, p_col="CTDPRS"):
    # TODO: import p_col from config file
    """
    Calculate pressure offset using deck pressure log and apply it to the data.
    Pressure flag column is added with value 2, indicating the data are calibrated.

    Parameters
    ----------
    df : DataFrame
        DataFrame containing column with pressure values
    p_col : str, optional
        Pressure column name in DataFrame (defaults to CTDPRS)

    Returns
    -------
    df : DataFrame
        DataFrame containing updated pressure values and a new flag column

    """
    p_log = pd.read_csv(
        cfg.dirs["logs"] + "ondeck_pressure.csv",
        dtype={"SSSCC": str},
        na_values="Started in Water",
    )
    p_offset = _get_pressure_offset(p_log.ondeck_start_p, p_log.ondeck_end_p)
    df[p_col] += p_offset
    df[p_col + "_FLAG_W"] = 2

    return df


def make_depth_log(time_df, threshold=80):
    # TODO: get column names from config file
    """
    Create depth log file from maximum depth of each station/cast in time DataFrame.
    If rosette does not get within the threshold distance of the bottom, returns NaN.

    Parameters
    ----------
    time_df : DataFrame
        DataFrame containing continuous CTD data
    threshold : int, optional
        Maximum altimeter reading to consider cast "at the bottom" (defaults to 80)

    """
    # TODO: make inputs be arraylike rather than dataframe
    df = time_df[["SSSCC", "CTDPRS", "GPSLAT", "ALT"]].copy().reset_index()
    df_group = df.groupby("SSSCC", sort=False)
    idx_p_max = df_group["CTDPRS"].idxmax()
    bottom_df = pd.DataFrame(
        data={
            "SSSCC": df["SSSCC"].unique(),
            "max_p": df.loc[idx_p_max, "CTDPRS"],
            "lat": df.loc[idx_p_max, "GPSLAT"],
            "alt": df.loc[idx_p_max, "ALT"],
        }
    )
    bottom_df.loc[bottom_df["alt"] > threshold, "alt"] = np.nan
    # pandas 1.2.1 ufunc issue workaround with pd.to_numpy()
    bottom_df["DEPTH"] = (
        (
            bottom_df["alt"]
            + np.abs(gsw.z_from_p(bottom_df["max_p"], bottom_df["lat"].to_numpy()))
        )
        .fillna(value=-999)
        .round()
        .astype(int)
    )
    bottom_df[["SSSCC", "DEPTH"]].to_csv(
        cfg.dirs["logs"] + "depth_log.csv", index=False
    )

    return True


def make_ssscc_list(fname="data/ssscc.csv"):
    """
    Attempt to automatically generate list of station/casts from raw files.
    """
    raw_files = Path(cfg.dirs["raw"]).glob("*.hex")
    ssscc_list = sorted([f.stem for f in raw_files])
    pd.Series(ssscc_list).to_csv(fname, header=None, index=False, mode="x")

    return ssscc_list


def get_ssscc_list(fname="data/ssscc.csv"):
    """
    Load in list of stations/casts to process.
    """
    ssscc_list = []
    with open(fname, "r") as filename:
        ssscc_list = [line.strip() for line in filename]

    return ssscc_list


def load_all_ctd_files(ssscc_list):
    """
    Load CTD files for station/cast list and merge into a dataframe.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to load

    Returns
    -------
    df_data_all : DataFrame
        Merged dataframe containing all loaded data

    """
    df_list = []
    for ssscc in ssscc_list:
        log.info("Loading TIME data for station: " + ssscc + "...")
        time_file = cfg.dirs["time"] + ssscc + "_time.pkl"
        time_data = pd.read_pickle(time_file)
        time_data["SSSCC"] = str(ssscc)
        time_data["dv_dt"] = oxy_fitting.calculate_dV_dt(
            time_data["CTDOXYVOLTS"], time_data["scan_datetime"]
        )
        df_list.append(time_data)
        # print("** Finished TIME data station: " + ssscc + " **")
    df_data_all = pd.concat(df_list, axis=0, sort=False)

    df_data_all["master_index"] = range(len(df_data_all))

    return df_data_all


def manual_backfill(df, p_cutoff, p_col="CTDPRS", flag_suffix="_FLAG_W"):
    """
    Overwrite values below cutoff pressure by backfilling the first data point past
    threshold upward to the surface. Backfilled data are flagged 6.

    Parameters
    ----------
    df : DataFrame
        Input data
    p_cutoff : float
        Cutoff pressure for backfilling
    p_col : str, optional
        Name of pressure column in df
    flag_suffix : str, optional
        Parameter suffix for data flags

    Returns
    -------
    df : DataFrame
        Input DataFrame with backfilled data
    """
    df = df.copy(deep=True)
    cols = df.columns.drop(p_col)
    df["interp_bool"] = df[p_col] < p_cutoff
    df.loc[df["interp_bool"], cols] = np.nan
    df = _flag_backfill_data(df).drop(columns="interp_bool")

    return df.bfill()


def _flag_backfill_data(
    df, p_col="CTDPRS", flag_bool_col="interp_bool", flag_suffix="_FLAG_W"
):
    """Flag data columns which have been interpolated with flag 6."""
    for col in df.columns:
        if flag_suffix in col:
            df.loc[df[flag_bool_col], col] = 6

    return df


def export_ct1(df, ssscc_list):
    """
    Export continuous CTD (i.e. time) data to data/pressure/ directory as well as
    adding quality flags and removing unneeded columns.

    Parameters
    ----------
    df : DataFrame
        Continuous CTD data
    ssscc_list : list of str
        List of stations to export

    Returns
    -------

    Notes
    -----
    Needs depth_log.csv and manual_depth_log.csv to run successfully

    """
    log.info("Exporting CTD files")

    # initial flagging (some of this should be moved)
    # TODO: lump all uncalibrated together; smart flagging like ["CTD*_FLAG_W"] = 1
    # TODO: may not always have these channels so don't hardcode them in!
    df["CTDFLUOR_FLAG_W"] = 1
    df["CTDXMISS_FLAG_W"] = 1
    # df["CTDBACKSCATTER_FLAG_W"] = 1

    # rename outputs as defined in user_settings.yaml
    for param, attrs in cfg.ctd_outputs.items():
        if param not in df.columns:
            df.rename(columns={attrs["sensor"]: param}, inplace=True)

    # check that all columns are there
    try:
        df[cfg.ctd_col_names]
        # this is lazy, do better
    except KeyError as err:
        log.info("Column names not configured properly... attempting to correct")
        bad_cols = err.args[0].split("'")[1::2]  # every other str is a column name
        for col in bad_cols:
            if col.endswith("FLAG_W"):
                log.warning(col + " missing, flagging with 9s")
                df[col] = 9
            else:
                log.warning(col + " missing, filling with -999s")
                df[col] = -999

    df["SSSCC"] = df["SSSCC"].astype(str).copy()
    cast_details = pd.read_csv(
        # cfg.dirs["logs"] + "cast_details.csv", dtype={"SSSCC": str}
        cfg.dirs["logs"] + "bottom_bottle_details.csv",
        dtype={"SSSCC": str},
    )
    depth_df = pd.read_csv(
        cfg.dirs["logs"] + "depth_log.csv", dtype={"SSSCC": str}, na_values=-999
    ).dropna()
    try:
        manual_depth_df = pd.read_csv(
            cfg.dirs["logs"] + "manual_depth_log.csv", dtype={"SSSCC": str}
        )
    except FileNotFoundError:
        # TODO: add logging; look into inheriting/extending a class to add features
        log.warning("manual_depth_log.csv not found... duplicating depth_log.csv")
        manual_depth_df = depth_df.copy()  # write manual_depth_log as copy of depth_log
        manual_depth_df.to_csv(cfg.dirs["logs"] + "manual_depth_log.csv", index=False)
    full_depth_df = pd.concat([depth_df, manual_depth_df])
    full_depth_df.drop_duplicates(subset="SSSCC", keep="first", inplace=True)

    for ssscc in ssscc_list:

        time_data = df[df["SSSCC"] == ssscc].copy()
        time_data = pressure_sequence(time_data)
        # switch oxygen primary sensor to rinko
        # if int(ssscc[:3]) > 35:
        print(f"Using Rinko as CTDOXY for {ssscc}")
        time_data.loc[:, "CTDOXY"] = time_data["CTDRINKO"]
        time_data.loc[:, "CTDOXY_FLAG_W"] = time_data["CTDRINKO_FLAG_W"]
        time_data = time_data[cfg.ctd_col_names]
        # time_data = time_data.round(4)
        time_data = time_data.where(~time_data.isnull(), -999)  # replace NaNs with -999

        # force flags back to int (TODO: make flags categorical)
        for col in time_data.columns:
            if col.endswith("FLAG_W"):
                time_data[col] = time_data[col].astype(int)

        try:
            depth = full_depth_df.loc[full_depth_df["SSSCC"] == ssscc, "DEPTH"].iloc[0]
        except IndexError:
            log.warning(f"No depth logged for {ssscc}, setting to -999")
            depth = -999

        # get cast_details for current SSSCC
        cast_dict = cast_details[cast_details["SSSCC"] == ssscc].to_dict("records")[0]
        b_datetime = (
            datetime.fromtimestamp(cast_dict["bottom_time"], tz=timezone.utc)
            .strftime("%Y%m%d %H%M")
            .split(" ")
        )
        # TODO: yo-yo casts are an edge case where this may be different
        btm_lat = cast_dict["latitude"]
        btm_lon = cast_dict["longitude"]

        now = datetime.now(timezone.utc)
        file_datetime = now.strftime("%Y%m%d")  # %H:%M")
        file_datetime = file_datetime + "ODFSIO"
        # TODO: only "cast" needs to be int; "station" is explicitly allowed to incl.
        # letters/etc. Moving from SSSCC to station & cast fields will be beneficial
        with open(f"{cfg.dirs['pressure']}{ssscc}_ct1.csv", "w+") as f:
            # put in logic to check columns?
            # number_headers should be calculated, not defined
            ctd_header = (  # this is ugly but prevents tabs before label
                f"CTD,{file_datetime}\n"
                f"NUMBER_HEADERS = 11\n"
                f"EXPOCODE = {cfg.expocode}\n"
                f"SECT_ID = {cfg.section_id}\n"
                f"STNNBR = {ssscc[:3]}\n"  # STNNBR = SSS
                f"CASTNO = {ssscc[3:]}\n"  # CASTNO = CC
                f"DATE = {b_datetime[0]}\n"
                f"TIME = {b_datetime[1]}\n"
                f"LATITUDE = {btm_lat:.4f}\n"
                f"LONGITUDE = {btm_lon:.4f}\n"
                f"INSTRUMENT_ID = {cfg.ctd_serial}\n"
                f"DEPTH = {depth:.0f}\n"
            )
            f.write(ctd_header)
            np.asarray(cfg.ctd_col_names).tofile(f, sep=",", format="%s")
            f.write("\n")
            np.asarray(cfg.ctd_col_units).tofile(f, sep=",", format="%s")
            f.write("\n")
            time_data.to_csv(f, header=False, index=False)
            f.write("END_DATA")



import matplotlib

from . import ctd_plots as ctd_plots
from . import merge_codes as merge

matplotlib.use("ps")

expocode = "320620180309"

qual_codes_filepath = "data/quality_codes/"
log_dir = "data/logs/quality_code/all/"

file_ssscc = "data/ssscc.csv"
bottle_file = f"{qual_codes_filepath}{expocode}_hy1.csv"

### compile bottle trips into one file, then merge with odf_db bottle file
df_bottle = merge.prelim_ctd_bottle_df(file_ssscc, bottle_file)
### compute residuals for plotting
merge.ctd_residuals_df(df_bottle)
df_bottle = merge.merge_ctd_codes(log_dir, df_bottle)
df_coded = merge.get_qc_all_from_df(df_bottle)
print(merge.residual_stddev(df_coded))

##### PATCHED IN FOR NBP1802 #####
df_coded = df_coded.rename(index=str, columns={"CTDCOND2_FLAG_W": "CTDSAL_FLAG_W"})
df_coded.to_pickle("data/values_and_codes.pkl")
# print(df_coded.axes)
##### PATCHED IN FOR NBP1802 #####

ctd_plots.make_cruise_report_plots(df_coded)
